(* Copyright (c) 2020 CIIRC (Czech Institute of Informatics, Robotics and Cybernetics) / CTU (Czech Technical University) *)

open Syntax
open Parser
open Interpret

let reporteachitem : bool ref = ref false
let reportids : bool ref = ref true
let reportpfcomplexity : bool ref = ref false
let webout : bool ref = ref false
let ajax : bool ref = ref false
let ajaxactive : bool ref = ref false
let ajaxpffile : string ref = ref ""
let sqlout : bool ref = ref false
let sqltermout : bool ref = ref false
let presentationonly : bool ref = ref false
let mainfilehash : string option ref = ref None
let solvesproblemfile : string option ref = ref None

let thmsasexercises : bool ref = ref false
let exercises : string list ref = ref []

let admits : bool ref = ref false

let indoutfile : string option ref = ref None;;
let indextms : (string,tp) Hashtbl.t = Hashtbl.create 1000;;
let indexknowns : (string,unit) Hashtbl.t = Hashtbl.create 1000;;
let ownedoutfile : string option ref = ref None;;
let ownedobj : (Hash.hashval,unit) Hashtbl.t = Hashtbl.create 1000;;
let ownedprop : (Hash.hashval,unit) Hashtbl.t = Hashtbl.create 1000;;

let combsigtm sgtmloc secstack =
  (List.concat
     (List.map (fun l -> List.map (fun (x,m,a,n,b) -> (x,(n,b))) l)
	(sgtmloc::(List.map (fun (_,_,_,stl,_) -> stl) secstack))))

exception AdmittedPf

let rec map_for f i n =
  if i <= n then
    (f i::map_for f (i+1) n)
  else
    [];;

let rec split_list n l =
  if n > 0 then
    match l with
    | (a::r) ->
      let (l1,l2) = split_list (n-1) r in
      (a::l1,l2)
    | [] -> ([],[])
  else
    ([],l)

let extract_pfg_id l =
  let ll = String.length l in
  let i = ref 0 in
  let spc = ref 0 in
  let rv = ref None in
  while !i < ll && !rv = None do
    let c = l.[!i] in
    incr i;
    if c = ' ' then incr spc;
    if !spc = 3 then rv := Some(String.sub l !i 64)
  done;
  begin
    match !rv with
    | Some(r) -> Hash.hexstring_hashval r
    | None -> raise (Failure "bad owned file line")
  end

let read_ownedfile c =
  try
    while true do
      let l = input_line c in
      if String.length l >= 68 && String.sub l 0 4 = "Obj " then
        Hashtbl.add ownedobj (extract_pfg_id l) ()
      else if String.length l >= 69 && String.sub l 0 5 = "Prop " then
        Hashtbl.add ownedprop (extract_pfg_id l) ()
    done
  with End_of_file -> ()

let read_indexfile c =
  let tl = ref (TokStrRest(Lexer.token,Lexing.from_channel c)) in
  lineno := 1;
  charno := 0;
  try
    while true do
      let (iitem,tr) = parse_indexitem !tl in
      tl := tr;
      match iitem with
      | IndexTm(h,a) ->
	  if not (valid_id_p h) then raise (Failure(h ^ " in index file is not a valid id"));
	  let a = ltree_to_atree a in
	  let atp = extract_tp a [] in
          if !verbosity > 10 then Printf.printf "  Hashtbl.add indextms \"%s\" (%s);\n" h (tp_ocaml atp);
	  begin
	    try
	      let itp = Hashtbl.find indextms h in
	      if itp <> atp then raise (Failure("Mismatch in indexed type assignment for " ^ h))
	    with Not_found ->
	      Hashtbl.add indextms h atp
	  end
      | IndexKnown(h) ->
	  if not (valid_id_p h) then raise (Failure(h ^ " in index file is not a valid id"));
          if !verbosity > 10 then Printf.printf "  Hashtbl.add indexknowns \"%s\" ();\n" h;
	  if not (Hashtbl.mem indexknowns h) then
	    Hashtbl.add indexknowns h ()
    done
  with
  | Lexer.Eof ->
      ()
  | Failure(x) ->
      if !webout then
	begin
          Printf.printf "AF%d:%d\n"  !lineno !charno;
	  Printf.printf "<div class='documentfail'>Failure reading index file at line %d char %d: %s</div>" !lineno !charno x; flush stdout;
	  exit 1
	end
      else if !ajax then
	begin
	  Printf.printf "f\n" (*** this indicates a fundamental problem, not a problem with the ajax input ***)
	end
      else
	begin
	  Printf.printf "Failure reading index file at line %d char %d: %s\n" !lineno !charno x; flush stdout;
	  exit 1
	end

let latex = ref None;;
let inchan = ref None;; (*** This is a possible second channel for reading the input file currently used to record a literal copy of the text of proofs ***)
let inchanline = ref 1;;
let inchanchar = ref 0;;
let includingsigfile = ref false;;
let includedsigfiles = ref [];;
let sigoutfile = ref None;;
let pfgout = ref false;;
type pfgitem = PfgParam of string * string * tp | PfgDef of string * string * tp * tm | PfgKnown of string * string * tm | PfgThm of string * string * tm * pf | PfgConj of string * string * tm
let pfgmain : pfgitem list ref = ref [];;
let pfgdelta : (string,unit) Hashtbl.t = Hashtbl.create 100;;
let pfgtmph : (string,string * tp * (tm option)) Hashtbl.t = Hashtbl.create 100;;
let pfgknph : (string,tm) Hashtbl.t = Hashtbl.create 100;;
let sigtmh : (string,string) Hashtbl.t  = Hashtbl.create 1000;;
let sigknh : (string,string) Hashtbl.t  = Hashtbl.create 1000;;
let sigtmof : (string,ptp) Hashtbl.t  = Hashtbl.create 1000;;
let sigdelta : (string,ptm) Hashtbl.t = Hashtbl.create 1000;;
let sigtm = ref [];;
let sigpf = ref [];;
let polytm = ref [];;
let polypf = ref [];;
let futurepolytm = ref [];;
let futurepolypf = ref [];;
let handlepolysnow () =
  polytm := !futurepolytm @ !polytm;
  polypf := !futurepolypf @ !polypf;
  futurepolytm := [];
  futurepolypf := [];;
let pushpolytm a = futurepolytm := a::!futurepolytm;;
let pushpolypf a = futurepolypf := a::!futurepolypf;;
(*** special cases so that the certain tactics get activated ***)
let activate_special_knowns h =
  if h = !expolyI then
    begin
      expolyIknown := true;
      if !verbosity > 50 then (Printf.printf "Activated expolyI, so that witness tactic can now be used.\n"; flush stdout);
    end;;
(*** I could make this more general, but I won't for now. ***)
let extract_exclaim m =
  match m with
  | Ap(TpAp(TmH(e),etp),ep) when e = !expoly -> (etp,ep)
  | _ -> raise Not_found

let add_sigdelta h (i,m) =
  let m = tm_beta_eta_norm m in
  match m with
  | TmH(_) -> ()
  | TpAp(TmH(_),TpVar(0)) -> ()
  | TpAp(TpAp(TmH(_),TpVar(0)),TpVar(1)) -> ()
  | TpAp(TpAp(TpAp(TmH(_),TpVar(0)),TpVar(1)),TpVar(2)) -> ()
  | _ -> Hashtbl.add sigdelta h (i,m)

(*** Output docitems to the signature file, changing theorems to axioms ***)
let outtosigfile soc ditem =
  let outc s = output_char soc s in
  let outs s = output_string soc s in
  let outi i = Printf.fprintf soc "%d" i in
  let outl a = output_ltree soc a in
  let outasckd asck =
    begin
      match asck with
      | AscTp -> outs " : ";
      | AscSet -> outs " :e ";
      | AscSubeq -> outs " c= ";
    end;		    
  in
  begin
    match ditem with
    | Section(x) ->
	outs ("Section " ^ x ^ ".\n")
    | End(x) ->
	outs ("End " ^ x ^ ".\n")
    | VarDecl(xl,asck,a) ->
	outs "Variable";
	List.iter (fun x -> outc ' '; outs x) xl;
	outasckd asck;
	outl a;
	outc '.';
	outc '\n';
    | HypDecl(x,a) ->
	outs ("Hypothesis " ^ x ^ " : ");
	outl a;
	outc '.';
	outc '\n';
    | LetDecl(x,None,b) ->
	outs ("Let " ^ x ^ " := ");
	outl b;
	outc '.';
	outc '\n';
    | LetDecl(x,Some(asck,a),b) ->
	outs "Let ";
	outs x;
	outasckd asck;
	outl a;
	outs " := ";
	outl b;
	outc '.';
	outc '\n';
    | PostInfixDecl(x,a,p,Postfix) ->
	outs "Postfix ";
	outs x;
	outc ' ';
	outi p;
	outs " := ";
	outl a;
	outs ".\n";
    | PostInfixDecl(x,a,p,InfixNone) ->
	outs "Infix ";
	outs x;
	outc ' ';
	outi p;
	outs " := ";
	outl a;
	outs ".\n";
    | PostInfixDecl(x,a,p,InfixLeft) ->
	outs "Infix ";
	outs x;
	outc ' ';
	outi p;
	outs " left ";
	outs " := ";
	outl a;
	outs ".\n";
    | PostInfixDecl(x,a,p,InfixRight) ->
	outs "Infix ";
	outs x;
	outc ' ';
	outi p;
	outs " right ";
	outs " := ";
	outl a;
	outs ".\n";
    | PrefixDecl(x,a,p) ->
	outs "Prefix ";
	outs x;
	outc ' ';
	outi p;
	outs " := ";
	outl a;
	outs ".\n";
    | BinderDecl(plus,comma,x,a,bo) ->
	outs "Binder";
	if plus then outs "+ " else outc ' ';
	outs x;
	if comma then outs " , := " else outs " => := ";
	outl a;
	begin
	  match bo with
	  | Some(b) -> outs " ; "; outl b
	  | None -> ()
	end;
	outs ".\n"
    | UnicodeDecl(x,yl) ->
	outs "(* Unicode ";
	outs x;
	List.iter (fun y ->
	  outs " \"";
	  outs y;
	  outs "\"")
	  yl;
	outs " *)\n"
    | SubscriptDecl(x) ->
	outs "(* Subscript ";
	outs x;
	outs " *)\n"
    | SuperscriptDecl(x) ->
	outs "(* Superscript ";
	outs x;
	outs " *)\n"
    | NotationDecl(x,yl) ->
	outs "Notation ";
	outs x;
	List.iter (fun y -> outc ' '; outs y) yl;
	outs ".\n";
    | ParamHash(x,h,None) ->
	outs "(* Parameter ";
	outs x;
	outc ' ';
	outc '"';
	outs h;
	outc '"';
	outc ' ';
	outs "*)\n";
    | ParamHash(x,h,Some(k)) ->
	outs "(* Parameter ";
	outs x;
	outc ' ';
	outc '"';
	outs h;
	outc '"';
	outc ' ';
	outc '"';
	outs k;
	outc '"';
	outc ' ';
	outs "*)\n";
    | ParamDecl(x,a) ->
	outs "Parameter ";
	outs x;
	outs " : ";
	outl a;
	outs ".\n";
    | DefDecl(x,None,b) ->
	outs "Definition ";
	outs x;
	outs " := ";
	outl b;
	outs ".\n";
    | DefDecl(x,Some a,b) ->
	outs "Definition ";
	outs x;
	outs " : ";
	outl a;
	outs " := ";
	outl b;
	outs ".\n";
    | AxDecl(x,a) ->
	outs "Axiom ";
	outs x;
	outs " : ";
	outl a;
	outs ".\n";
    | ThmDecl(_,x,a) -> (*** Theorems are axioms in the signature. ***)
	outs "Axiom ";
	outs x;
	outs " : ";
	outl a;
	outs ".\n";
    | _ -> ()
  end

let skip_to_line_char c li1 ch1 li2 ch2 =
  try
    while (!li1 < li2) do
      ch1 := 0;
      let z = input_char c in
      if z = '\n' then incr li1;
    done;
    while (!ch1 < ch2) do
      ignore (input_char c);
      incr ch1
    done
  with End_of_file -> () (*** this probably shouldn't happen, but if it does, just stop reading the file ***)

let buffer_to_line_char c b li1 ch1 li2 ch2 =
  try
    while (!li1 < li2) do
      ch1 := 0;
      let z = input_char c in
      Buffer.add_char b z;
      if z = '\n' then incr li1;
    done;
    while (!ch1 < ch2) do
      let z = input_char c in
      Buffer.add_char b z;
      incr ch1
    done
  with End_of_file -> () (*** this probably shouldn't happen, but if it does, just stop reading the file ***)

let ctxtp = ref []
let ctxtm = ref []
let ctxpf = ref []
let tparclos = ref (fun a -> a)
let tmallclos = ref (fun m -> m)
let tmlamclos = ref (fun m -> m)
let pflamclos = ref (fun d -> d)
let aptmloc = ref (fun m cxtp cxtm -> m)
let appfloc = ref (fun d cxtp cxtm cxpf -> d)
let secstack = ref []
let popfn = ref (fun () -> ())

let laststructaction = ref 0;;

let evaluate_docitem_1 ditem =
  begin
    match !sigoutfile with
    | Some soc -> outtosigfile soc ditem
    | None -> ()
  end;
  match ditem with
  | Author(x,yl) ->
      authors := !authors @ (x::yl);
      begin
	List.iter
	  (fun z ->
	    if !sqlout then
	      begin
		match !mainfilehash with
		| Some docsha ->
		    if not !presentationonly then (Printf.printf "INSERT INTO `docauthor` (`docsha`,`docauthorname`) VALUES ('%s',\"%s\");\n" docsha (String.escaped z));
		| None -> ()
	      end)
	  (x::yl);
      end
  | Title(x) ->
      begin
	match !title with
	| None -> title := Some(x);
	    if !sqlout then
	      begin
		match !mainfilehash with
		| Some docsha ->
		    if not !presentationonly then (Printf.printf "INSERT INTO `doc` (`docsha`,`docname`,`doctitle`,`dockind`,`docreleaseorder`,`docreleasedate`,`docstatus`) VALUES ('%s','%s',\"%s\",'putkind',1,'putrdate','s');\n" docsha Sys.argv.((Array.length Sys.argv) - 1) (String.escaped x));
		| None -> ()
	      end;
	| Some _ -> raise (Failure("Title can only be declared once"))
      end
  | ShowProofTerms(b) -> showproofterms := b
  | Salt(x) -> salt := Some x
  | Treasure(x) ->
      treasure := Some x;
  | Section(x) ->
      begin
	secstack := (x,!popfn,!aptmloc,!appfloc,!sigtm,!sigpf)::!secstack;
	popfn := (fun () -> ());
      end
  | End(x) ->
      begin
	match !secstack with
	| ((y,f,atl,apl,st,sp)::cs) when x = y ->
	    begin
	      !popfn ();
	      popfn := f;
	      aptmloc := atl;
	      appfloc := apl;
	      sigtm := st;
	      sigpf := sp;
	      secstack := cs
	    end
	| ((y,f,atl,apl,stl,spl)::_) -> raise (Failure("Section " ^ y ^ " cannot be ended with End " ^ x))
	| [] -> raise (Failure("No Section To End"))
      end
  | VarDecl(xl,AscTp,NaL "SType") -> (** Type variable **)
      begin
	match (!ctxtp,!ctxtm,!ctxpf) with
	| ([],[],[]) ->
	    ctxtp := xl;
	    if (List.length !ctxtp > 6) then raise (Failure "More than 6 type variables are not allowed.");
	    List.iter (fun x ->
	      let prevaptmloc = !aptmloc in
	      let prevappfloc = !appfloc in
	      aptmloc := (fun m cxtp cxtm -> TpAp(prevaptmloc m cxtp cxtm,tplookup cxtp x));
	      appfloc := (fun d cxtp cxtm cxpf -> PTpAp(prevappfloc d cxtp cxtm cxpf,tplookup cxtp x)))
	      xl;
	    let popfnnow = !popfn in
	    popfn := (fun () -> handlepolysnow (); ctxtp := []; popfnnow ())
	| _ ->
	    raise (Failure "Type variables can only be declared when the context is empty.")
      end
  | VarDecl(xl,AscTp,a) -> (** Ordinary variable **)
      let a = ltree_to_atree a in
      let atp = extract_tp a !ctxtp in
      let prevctxtm = !ctxtm in
      let prevctxpf = !ctxpf in
      let prevtparclos = !tparclos in
      let prevtmallclos = !tmallclos in
      let prevtmlamclos = !tmlamclos in
      let prevpflamclos = !pflamclos in
      ctxtm := (List.map (fun x -> (x,(atp,None))) (List.rev xl)) @ prevctxtm;
      let xln = List.length xl in
      ctxpf := (List.map (fun (y,p) -> (y,tmshift 0 xln p)) prevctxpf); (*** shift the hyp context to account for the fresh vars on the var context ***)
      List.iter (fun x ->
	let prevaptmloc = !aptmloc in
	let prevappfloc = !appfloc in
	aptmloc := (fun m cxtp cxtm -> Ap(prevaptmloc m cxtp cxtm,tmlookup cxtm x));
	appfloc := (fun d cxtp cxtm cxpf -> PTmAp(prevappfloc d cxtp cxtm cxpf,tmlookup cxtm x)))
	xl;
      tparclos := (fun a -> prevtparclos (List.fold_right (fun x b -> Ar(atp,b)) xl a));
      tmallclos := (fun m -> prevtmallclos (List.fold_right (fun x n -> All(atp,n)) xl m));
      tmlamclos := (fun m -> prevtmlamclos (List.fold_right (fun x n -> Lam(atp,n)) xl m));
      pflamclos := (fun d -> prevpflamclos (List.fold_right (fun x e -> TLam(atp,e)) xl d));
      let popfnnow = !popfn in
      popfn := (fun () -> ctxtm := prevctxtm; ctxpf := prevctxpf; tparclos := prevtparclos; tmallclos := prevtmallclos; tmlamclos := prevtmlamclos; pflamclos := prevpflamclos; popfnnow ())
  | VarDecl(xl,AscSet,a) -> (** Variable/Hyp combo, I might allow this later **)
      raise (Failure("Variables must be ascribed a type, not a set"))
  | VarDecl(xl,AscSubeq,a) -> (** Variable/Hyp combo, I might allow this later **)
      raise (Failure("Variables must be ascribed a type, not a set"))
  | HypDecl(x,a) -> (** declare a hypothesis, put into ctxpf, add removal to popfn, update appfloc **)
      let a = ltree_to_atree a in
      let atm = check_tm a Prop !polytm sigtmof !sigtm !ctxtp !ctxtm in
      let prevappfloc = !appfloc in
      let prevtmallclos = !tmallclos in
      let prevpflamclos = !pflamclos in
      let prevctxpf = !ctxpf in
      ctxpf := (x,atm)::prevctxpf;
      appfloc := (fun d cxtp cxtm cxpf -> PPfAp(prevappfloc d cxtp cxtm cxpf,pflookup cxpf x));
      tmallclos := (fun m -> prevtmallclos (Imp(atm,m)));
      pflamclos := (fun d -> prevpflamclos (PLam(atm,d)));
      let popfnnow = !popfn in
      popfn := (fun () -> ctxpf := prevctxpf; tmallclos := prevtmallclos; pflamclos := prevpflamclos; popfnnow ())
  | LetDecl(x,None,b) -> (** add (x,atp[extracted from btm],btm) to ctxtm, add removal to popfn **)
      (*** Note: Let variables do not correspond to de Bruijn indices. There need be no shifting of hypotheses. ***)
      let b = ltree_to_atree b in
      let (btm,btp) = extract_tm b !polytm sigtmof !sigtm !ctxtp !ctxtm in
      let prevctxtm = !ctxtm in
      ctxtm := (x,(btp,Some btm))::prevctxtm;
      let popfnnow = !popfn in
      popfn := (fun () -> ctxtm := prevctxtm; popfnnow ())
  | LetDecl(x,Some (AscTp,a),b) -> (** add (x,atp,btm) to ctxtm, add removal to popfn **)
      let a = ltree_to_atree a in
      let b = ltree_to_atree b in
      let btp = extract_tp a !ctxtp in
      let btm = check_tm b btp !polytm sigtmof !sigtm !ctxtp !ctxtm in
      let prevctxtm = !ctxtm in
      ctxtm := (x,(btp,Some btm))::prevctxtm;
      let popfnnow = !popfn in
      popfn := (fun () -> ctxtm := prevctxtm; popfnnow ())
  | LetDecl(x,Some (_,a),b) -> (*** I doubt I will allow this later since it would require finding a proof that b is in the set a ***)
      raise (Failure("Lets must be ascribed a type, not a set"))
  | PostInfixDecl(x,a,p,pic) ->
      let a = ltree_to_atree a in
      popfn := declare_postinfix x a p pic !popfn
  | PrefixDecl(x,a,p) ->
      let a = ltree_to_atree a in
      popfn := declare_prefix x a p !popfn
  | BinderDecl(plus,comma,x,a,bo) ->
      let a = ltree_to_atree a in
      let bo = (match bo with Some(b) -> Some(ltree_to_atree b) | None -> None) in
      popfn := declare_binder plus comma x a bo !popfn
  | UnicodeDecl(x,ul) ->
      Hashtbl.add unicode x ul; (*** just associate x with ul forever - not contained within a section ***)
  | SubscriptDecl(x) -> Hashtbl.add subscript x () (*** escapes sections, only use this when it should be permanent ***)
  | SuperscriptDecl(x) -> Hashtbl.add superscript x () (*** escapes sections, only use this when it should be permanent ***)
  | NotationDecl(x,yl) ->
      begin
	match x with
	| "IfThenElse" ->
	    begin
	      match yl with
	      | [y] ->
		  begin
		    try
		      let h = Hashtbl.find sigtmh y in
		      match Hashtbl.find sigtmof h with
		      | (1,Ar(Prop,Ar(TpVar(0),Ar(TpVar(0),TpVar(0))))) ->
			 ifop := Some(h)
                      | (0,Ar(Prop,Ar(Set,Ar(Set,Set)))) ->
                         ifopset := Some(h)
		      | _ -> raise (Failure("IfThenElse Notation should be given a polymorphic name of type prop->?0->?0->?0 or prop->set->set->set"))
		    with Not_found -> raise (Failure("IfThenElse Notation should be given a polymorphic name of type prop->?0->?0->?0 or prop->set->set->set"))
		  end
  	      | _ -> raise (Failure("IfThenElse Notation should be given a polymorphic name of type prop->?0->?0->?0 or prop->set->set->set"))
	    end
	| "Repl" ->
	    begin
	      match yl with
	      | [y] ->
		  begin
		    try
		      let h = Hashtbl.find sigtmh y in
		      match Hashtbl.find sigtmof h with
		      | (0,Ar(Set,Ar(Ar(Set,Set),Set))) ->
			  replop := Some(h)
		      | _ -> raise (Failure("Repl Notation should be given a parameter or definition name of type set->(set->set)->set"))
		    with Not_found -> raise (Failure("Repl Notation should be given a parameter or definition name of type set->(set->set)->set"))
		  end
  	      | _ -> raise (Failure("Repl Notation should be given a parameter or definition name of type set->(set->set)->set"))
	    end
	| "Sep" ->
	    begin
	      match yl with
	      | [y] ->
		  begin
		    try
		      let h = Hashtbl.find sigtmh y in
		      match Hashtbl.find sigtmof h with
		      | (0,Ar(Set,Ar(Ar(Set,Prop),Set))) ->
			  sepop := Some(h)
		      | _ -> raise (Failure("Repl Notation should be given a parameter or definition name of type set->(set->set)->set"))
		    with Not_found -> raise (Failure("Repl Notation should be given a parameter or definition name of type set->(set->set)->set"))
		  end
  	      | _ -> raise (Failure("Repl Notation should be given a parameter or definition name of type set->(set->set)->set"))
	    end
	| "ReplSep" ->
	    begin
	      match yl with
	      | [y] ->
		  begin
		    try
		      let h = Hashtbl.find sigtmh y in
		      match Hashtbl.find sigtmof h with
		      | (0,Ar(Set,Ar(Ar(Set,Prop),Ar(Ar(Set,Set),Set)))) ->
			  replsepop := Some(h)
		      | _ -> raise (Failure("ReplSep Notation should be given a parameter or definition name of type set->(set->prop)->(set->set)->set"))
		    with Not_found -> raise (Failure("ReplSep Notation should be given a parameter or definition name of type set->(set->prop)->(set->set)->set"))
		  end
  	      | _ -> raise (Failure("ReplSep Notation should be given a parameter or definition name of type set->(set->prop)->(set->set)->set"))
	    end
	| "SetEnum" ->
	    begin
	      if List.length yl <= 1 then
		raise (Failure("SetEnum should be given at least n>=2 names: constructors for the first n-1 cases and finally constructor for adjoining to a set"))
	      else
		let rec f i yl ztp =
		  match yl with
		  | [y] ->
		      begin
			try
			  let h = Hashtbl.find sigtmh y in
			  match Hashtbl.find sigtmof h with
			  | (0,Ar(Set,Ar(Set,Set))) ->
			      setenumadj := Some(h)
			  | _ -> raise (Failure("The last name in the SetEnum Notation should be given a parameter or definition name of type set->set->set"))
			with Not_found -> raise (Failure("The last name in the SetEnum Notation should be given a parameter or definition name of type set->set->set"))
		      end
		  | (y::yr) ->
		      begin
			try
			  let h = Hashtbl.find sigtmh y in
			  match Hashtbl.find sigtmof h with
			  | (0,htp) ->
			      if htp = ztp then
				setenuml := Some(h)::!setenuml
			      else
				begin
				  Printf.printf "y = %s h = %s\nhtp = %s\nztp = %s\n" y h (tp_to_str htp) (tp_to_str ztp); flush stdout;
				  raise (Failure("Name " ^ string_of_int i ^ " in the SetEnum Notation should be given a parameter or definition name of type " ^ tp_to_str ztp))
				end
			  | _ -> raise (Failure("Name " ^ string_of_int i ^ " in the SetEnum Notation should be given a parameter or definition name of type " ^ tp_to_str ztp))
			with Not_found -> raise (Failure("Name " ^ string_of_int i ^ " in the SetEnum Notation should be given a parameter or definition name of type " ^ tp_to_str ztp))
		      end;
		      f (i+1) yr (Ar(Set,ztp))
		  | [] -> raise (Failure("SetEnum should be given at least n>=2 names: constructors for the first n-1 cases and finally constructor for adjoining to a set"))
		in
		setenuml := [];
		f 0 yl Set;
		setenuml := List.rev !setenuml
	    end
	| "SetEnum0" ->
	    begin
	      match yl with
	      | [y] ->
		  begin
		    try
		      let h = Hashtbl.find sigtmh y in
		      match Hashtbl.find sigtmof h with
		      | (0,Set) ->
			  set_setenuml_n 0 h
		      | _ -> raise (Failure("SetEnum0 Notation should be given a parameter or definition name of type set"))
		    with Not_found -> raise (Failure("SetEnum0 Notation should be given a parameter or definition name of type set"))
		  end
  	      | _ -> raise (Failure("SetEnum0 Notation should be given a parameter or definition name of type set"))
	    end
	| "SetEnum1" ->
	    begin
	      match yl with
	      | [y] ->
		  begin
		    try
		      let h = Hashtbl.find sigtmh y in
		      match Hashtbl.find sigtmof h with
		      | (0,Ar(Set,Set)) ->
			  set_setenuml_n 1 h
		      | _ -> raise (Failure("SetEnum1 Notation should be given a parameter or definition name of type set->set"))
		    with Not_found -> raise (Failure("SetEnum1 Notation should be given a parameter or definition name of type set->set"))
		  end
  	      | _ -> raise (Failure("SetEnum1 Notation should be given a parameter or definition name of type set->set"))
	    end
	| "SetEnum2" ->
	    begin
	      match yl with
	      | [y] ->
		  begin
		    try
		      let h = Hashtbl.find sigtmh y in
		      match Hashtbl.find sigtmof h with
		      | (0,Ar(Set,Ar(Set,Set))) ->
			  set_setenuml_n 2 h;
		      | _ -> raise (Failure("SetEnum2 Notation should be given a parameter or definition name of type set->set->set"))
		    with Not_found -> raise (Failure("SetEnum2 Notation should be given a parameter or definition name of type set->set->set"))
		  end
  	      | _ -> raise (Failure("SetEnum2 Notation should be given a parameter or definition name of type set->set->set"))
	    end
	| "Nat" ->
	    begin
	      match yl with
	      | [y0;yS] ->
		  begin
		    try
		      let h0 = Hashtbl.find sigtmh y0 in
		      let hS = Hashtbl.find sigtmh yS in
		      match (Hashtbl.find sigtmof h0,Hashtbl.find sigtmof hS) with
		      | ((0,Set),(0,Ar(Set,Set))) ->
			  nat0 := Some(h0);
			  natS := Some(hS)
		      | _ -> raise (Failure("Nat Notation should be given a name of type set (for 0) and a name of type set->set (for successor)"))
		    with Not_found -> raise (Failure("Nat Notation should be given a name of type set (for 0) and a name of type set->set (for successor)"))
		  end
	      | _ -> raise (Failure("Nat Notation should be given a name of type set (for 0) and a name of type set->set (for successor)"))
	    end
	| "SetLam" ->
	    begin
	      match yl with
	      | [y] ->
		  begin
		    try
		      let h = Hashtbl.find sigtmh y in
		      match Hashtbl.find sigtmof h with
		      | (0,Ar(Set,Ar(Ar(Set,Set),Set))) ->
			  setlam := Some h
		      | _ -> raise (Failure("SetLam Notation should be given a name of type set->(set->set)->set"))
		    with Not_found -> raise (Failure("SetLam Notation should be given a name of type set->(set->set)->set"))
		  end
	      | _ -> raise (Failure("SetLam Notation should be given a name of type set->(set->set)->set"))
	    end
	| "SetImplicitOp" ->
	    begin
	      match yl with
	      | [y] ->
		  begin
		    try
		      let h = Hashtbl.find sigtmh y in
		      match Hashtbl.find sigtmof h with
		      | (0,Ar(Set,Ar(Set,Set))) ->
			  setimplop := Some h
		      | _ -> raise (Failure("SetAp Notation should be given a name of type set->set->set"))
		    with Not_found -> raise (Failure("SetAp Notation should be given a name of type set->set->set"))
		  end
	      | _ -> raise (Failure("SetAp Notation should be given a name of type set->set->set"))
	    end
	| _ -> raise (Failure("Unknown Notation " ^ x))
      end
  | ParamHash(x,h,ok) ->
      if (!verbosity > 9) then (Printf.printf "ParamHash %s %s\n" x h; flush stdout);
      begin
	try
	  let (xj,(xi,_)) = primname x in
	  let xh = ptm_lam_id (xi,Prim xj) sigtmof sigdelta in
	  if (h <> xh) then raise (Failure(x ^ " is the name of a built-in primitive and must have the id " ^ xh))
	with
        | Not_found ->
           if Hashtbl.mem sigtmh x then raise (Failure(x ^ " is already assigned an id"));
           Hashtbl.add sigtmh x h;
           match ok with (** proofgold id **)
           | None -> ()
           | Some(k) -> Hashtbl.add pfgtmhh h (Hash.hexstring_hashval k)
      end
  | ParamDecl(x,a) ->
      if !reporteachitem then (Printf.printf "++ %s\n" x; flush stdout);
      let a = ltree_to_atree a in
      if List.mem_assoc x !sigtm || List.mem_assoc x !sigpf || List.mem x !ctxtp || List.mem_assoc x !ctxtm || List.mem_assoc x !ctxpf then
	raise (Failure(x ^ " has already been used."));
      if (!verbosity > 9) then (Printf.printf "Param %s : " x; output_ltree stdout (atree_to_ltree a); Printf.printf "\n"; flush stdout);
      let i = List.length !ctxtp in
      let atp = extract_tp a !ctxtp in
      if (!verbosity > 19) then Printf.printf "i = %d\natp = %s\n" i (tp_to_str atp);
      let agtp = !tparclos atp in
      begin
	try
	  let (xj,(xi,xtp)) = primname x in
	  if (xi,xtp) <> (i,agtp) then raise (Failure(x ^ " is the name of a built-in primitive which does not have the given type."));
	  if i > 6 then raise (Failure("It is forbidden to have more than 6 type variables."));
	  let xhv = tm_id (Prim(xj)) sigtmof sigdelta in
	  begin
	    if !sqlout then
	      begin
		match !mainfilehash with
		| Some docsha ->
		    if !sqltermout then Printf.printf "INSERT INTO `term` (`termid`,`termtp`,`termpoly`,`termprimitive`) VALUES ('%s','%s',%d,true);\n" xhv (stp_html_string agtp) i;
		    if not !presentationonly then (Printf.printf "INSERT INTO `termdoc` (`termid`,`docsha`,`termdocname`,`termdockind`) VALUES ('%s','%s',\"%s\",'p');\n" xhv docsha (String.escaped x));
		| None -> ()
	      end
	  end;
	  if !verbosity > 3 then (Printf.printf "%s has id %s\n" x xhv; flush stdout);
	  if (x = "Power") then setPow := Some xhv;
	  if (x = "In") then (*** As soon as In is declared, declare the id's corresponding to :e and c= ***)
	    begin
	      setIn := Some xhv;
	      let subhv = ptm_lam_id (0,Lam(Set,Lam(Set,All(Set,Imp(Ap(Ap(TmH(xhv),DB(0)),DB(2)),Ap(Ap(TmH(xhv),DB(0)),DB(1))))))) sigtmof sigdelta in
	      setSubeq := Some subhv
	    end;
	  Hashtbl.add sigtmh x xhv;
	  Hashtbl.add sigtmof xhv (i,agtp);
	  if i > 0 then (*** x will look polymorphic with i types after the appropriate section is ended ***)
	    pushpolytm ((x,i),agtp);
	  if i = 0 && not (Hashtbl.mem indextms xhv) then Hashtbl.add indextms xhv xtp; (*** since this is a primitive, it doesn't really need to be indexed. However, indexing it will allow me to use parameters with different names for the primitives if I want. ***)
	  let m = TmH xhv in
	  sigtm := (x,!aptmloc m)::!sigtm;
	  secstack := List.map (fun (y,f,atl,apl,st,sp) -> (y,f,atl,apl,(x,atl m)::st,sp)) !secstack
	with Not_found ->
	  begin
	    if (i > 0) then raise (Failure(x ^ " must be defined. The only polymorphic parameter allowed are built-in primitives."));
	    if List.mem_assoc x !sigtm || List.mem_assoc x !sigpf || List.mem x !ctxtp || List.mem_assoc x !ctxtm || List.mem_assoc x !ctxpf then
	      raise (Failure(x ^ " has already been used."))
	    else
	      try
		let xhv = Hashtbl.find sigtmh x in
		begin
                  if i = 0 then Hashtbl.add pfgtmph xhv (x,agtp,None);
		  if !pfgout && i = 0 then pfgmain := PfgParam(xhv,x,agtp)::!pfgmain;
		end;
		begin
		  if !sqlout then
		    begin
		      match !mainfilehash with
		      | Some docsha ->
			  if !sqltermout then Printf.printf "INSERT INTO `term` (`termid`,`termtp`,`termpoly`) VALUES ('%s','%s',%d);\n" xhv (stp_html_string agtp) i;
			  if not !presentationonly then (Printf.printf "INSERT INTO `termdoc` (`termid`,`docsha`,`termdocname`,`termdockind`) VALUES ('%s','%s',\"%s\",'p');\n" xhv docsha (String.escaped x));
		      | None -> ()
		    end
		end;
		begin
		  try
		    if (x = "Subq") then
		      begin
			match !setSubeq with
			| Some subhv ->
			    if xhv <> subhv then
			      raise (Failure("Subq can only be used to mean fun X Y:set => forall x:set, x :e X -> x :e Y"))
			| None ->
			    raise (Failure("Subq can only be used after In is declared and then only to mean fun X Y:set => forall x:set, x :e X -> x :e Y"))
		      end;
		    let itp = Hashtbl.find indextms xhv in (** megalodon knows it **)
		    if agtp <> itp then raise (Failure("The id " ^ xhv ^ " associated with the parameter " ^ x ^ " has is indexed to have the type " ^ tp_to_str itp ^ " not " ^ tp_to_str agtp));
		    Hashtbl.add sigtmof xhv (i,agtp);
		    let m = TmH xhv in
		    sigtm := (x,!aptmloc m)::!sigtm;
		    secstack := List.map (fun (y,f,atl,apl,st,sp) -> (y,f,atl,apl,(x,atl m)::st,sp)) !secstack
		  with
                  | Not_found ->
                     try
                       if i > 0 then raise Not_found;
                       let pfghpure = Hashtbl.find pfgtmhh xhv in
                       let pfghthy = pfg_objid_pure_to_thy pfghpure agtp in
                       if not (Hashtbl.mem ownedobj pfghthy) then raise Not_found; (** proofgold knows it **)
		       Hashtbl.add sigtmof xhv (i,agtp);
		       let m = TmH xhv in
		       sigtm := (x,!aptmloc m)::!sigtm;
		       secstack := List.map (fun (y,f,atl,apl,st,sp) -> (y,f,atl,apl,(x,atl m)::st,sp)) !secstack
		     with
                     | Not_found ->
		        raise (Failure("The given id " ^ xhv ^ " for " ^ x ^ " is not a known index for a term."))
		end
   	      with Not_found ->
		raise (Failure("Unknown id for " ^ x))
	  end
      end
  | DefDecl(x,None,b) ->
      if !reporteachitem then (Printf.printf "++ %s\n" x; flush stdout);
      let b = ltree_to_atree b in
      if List.mem_assoc x !sigtm || List.mem_assoc x !sigpf || List.mem x !ctxtp || List.mem_assoc x !ctxtm || List.mem_assoc x !ctxpf then
	raise (Failure(x ^ " has already been used."))
      else
	begin
	  if (!verbosity > 9) then (Printf.printf "Def %s := " x; output_ltree stdout (atree_to_ltree b); Printf.printf "\n"; flush stdout);
	  let i = List.length !ctxtp in
	  let (btm,btp) = extract_tm b !polytm sigtmof !sigtm !ctxtp !ctxtm in
	  if (!verbosity > 9) then (Printf.printf "Def %s := " x; output_ltree stdout (atree_to_ltree b); Printf.printf "\n"; flush stdout);
	  let bgtm = !tmlamclos btm in
	  let bgtp = !tparclos btp in
	  let xhv = ptm_lam_id (i,bgtm) sigtmof sigdelta in
          begin
            if i = 0 then
              let (pure,pfghv) = pfg_objid bgtm bgtp in
	      if (!verbosity > 3) then (Printf.printf "%s was assigned pfg obj id %s\n" x (Hash.hashval_hexstring pfghv); flush stdout);
              Hashtbl.add ownedobj pfghv ();
              if (!verbosity > 3) then (Printf.printf "(* Parameter %s \"%s\" \"%s\" *)\n" x xhv (Hash.hashval_hexstring pure));
              Hashtbl.add pfgtmhh xhv pure;
          end;
	  add_sigdelta xhv (i,bgtm);
	  begin
            if i = 0 then Hashtbl.add pfgtmph xhv (x,bgtp,Some(bgtm));
	    if !pfgout && i = 0 then pfgmain := PfgDef(xhv,x,bgtp,bgtm)::!pfgmain;
	  end;
	  begin
	    if !sqlout then
	      begin
		match !mainfilehash with
		| Some docsha ->
		    if !sqltermout then Printf.printf "INSERT INTO `term` (`termid`,`termtp`,`termpoly`) VALUES ('%s','%s',%d);\n" xhv (stp_html_string bgtp) i;
		    if not !presentationonly then (Printf.printf "INSERT INTO `termdoc` (`termid`,`docsha`,`termdocname`,`termdockind`) VALUES ('%s','%s',\"%s\",'d');\n" xhv docsha (String.escaped x));
		| None -> ()
	      end
	  end;
	  if (x = "Subq") then
	    begin
	      match !setSubeq with
	      | Some subhv ->
		  if xhv <> subhv then
		    raise (Failure("Subq can only be used to mean fun X Y:set => forall x:set, x :e X -> x :e Y"))
	      | None ->
		  raise (Failure("Subq can only be used after In is declared and then only to mean fun X Y:set => forall x:set, x :e X -> x :e Y"))
	    end;
	  if (!verbosity > 3) then (Printf.printf "%s := %s was assigned id %s\n" x (tm_to_str bgtm) xhv; flush stdout);
	  if i > 0 then (*** x will look polymorphic with i types after the appropriate section is ended ***)
	    pushpolytm ((x,i),bgtp);
	  begin
	    try
	      let xhv2 = Hashtbl.find sigtmh x in
	      if xhv <> xhv2 then raise (Failure(x ^ " was explicitly assigned id " ^ xhv2 ^ " but this does not match its computed id " ^ xhv))
	    with Not_found -> ()
	  end;
	  Hashtbl.add sigtmh x xhv;
	  Hashtbl.add sigtmof xhv (i,bgtp);
	  if i = 0 then
	    begin
	      try
		let itp = Hashtbl.find indextms xhv in
		if bgtp <> itp then raise (Failure("The id " ^ xhv ^ " associated with the definition of " ^ x ^ " has is indexed to have the type " ^ tp_to_str itp ^ " not " ^ tp_to_str bgtp ^ ". This indicates some fundamental bug since it would presumably imply a hash collision."));
	      with Not_found ->
		Hashtbl.add indextms xhv bgtp
	    end;
	  let m = TmH xhv in
	  sigtm := (x,!aptmloc m)::!sigtm;
	  secstack := List.map (fun (y,f,atl,apl,st,sp) -> (y,f,atl,apl,(x,atl m)::st,sp)) !secstack;
	  if (!verbosity > 19) then (Printf.printf "i = %d\nbtm = %s\nbtp = %s\n" i (tm_to_str btm) (tp_to_str btp); flush stdout);
	end
  | DefDecl(x,Some a,b) ->
      if !reporteachitem then (Printf.printf "++ %s\n" x; flush stdout);
      let a = ltree_to_atree a in
      let b = ltree_to_atree b in
      if List.mem_assoc x !sigtm || List.mem_assoc x !sigpf || List.mem x !ctxtp || List.mem_assoc x !ctxtm || List.mem_assoc x !ctxpf then
	raise (Failure(x ^ " has already been used."))
      else
	begin
	  if (!verbosity > 9) then (Printf.printf "Def %s : " x; output_ltree stdout (atree_to_ltree a); Printf.printf "\n := "; output_ltree stdout (atree_to_ltree b); Printf.printf "\n"; flush stdout);
	  let i = List.length !ctxtp in
	  let atp = extract_tp a !ctxtp in
	  let btm = check_tm b atp !polytm sigtmof !sigtm !ctxtp !ctxtm in
	  if (!verbosity > 9) then (Printf.printf "Def %s := " x; output_ltree stdout (atree_to_ltree b); Printf.printf "\n"; flush stdout);
	  let bgtm = !tmlamclos btm in
	  let agtp = !tparclos atp in
	  let xhv = ptm_lam_id (i,bgtm) sigtmof sigdelta in
          begin
            if i = 0 then
              let (pure,pfghv) = pfg_objid bgtm agtp in
	      if (!verbosity > 3) then (Printf.printf "%s was assigned pfg obj id %s\n" x (Hash.hashval_hexstring pfghv); flush stdout);
              Hashtbl.add ownedobj pfghv ();
              if (!verbosity > 3) then (Printf.printf "(* Parameter %s \"%s\" \"%s\" *)\n" x xhv (Hash.hashval_hexstring pure));
              Hashtbl.add pfgtmhh xhv pure;
          end;
	  add_sigdelta xhv (i,bgtm);
	  begin
            if i = 0 then Hashtbl.add pfgtmph xhv (x,agtp,Some(bgtm));
	    if !pfgout && i = 0 then pfgmain := PfgDef(xhv,x,agtp,bgtm)::!pfgmain;
	  end;
	  begin
	    if !sqlout then
	      begin
		match !mainfilehash with
		| Some docsha ->
		    if !sqltermout then Printf.printf "INSERT INTO `term` (`termid`,`termtp`,`termpoly`) VALUES ('%s','%s',%d);\n" xhv (stp_html_string agtp) i;
		    if not !presentationonly then (Printf.printf "INSERT INTO `termdoc` (`termid`,`docsha`,`termdocname`,`termdockind`) VALUES ('%s','%s',\"%s\",'d');\n" xhv docsha (String.escaped x));
		| None -> ()
	      end
	  end;
	  if (x = "Subq") then
	    begin
	      match !setSubeq with
	      | Some subhv ->
		  if xhv <> subhv then
		    raise (Failure("Subq can only be used to mean fun X Y:set => forall x:set, x :e X -> x :e Y"))
	      | None ->
		  raise (Failure("Subq can only be used after In is declared and then only to mean fun X Y:set => forall x:set, x :e X -> x :e Y"))
	    end;
	  if (!verbosity > 3) then (Printf.printf "%s := %s was assigned id %s\n" x (tm_to_str bgtm) xhv; flush stdout);
	  if i > 0 then (*** x will look polymorphic with i types after the appropriate section is ended ***)
	    pushpolytm ((x,i),agtp);
	  begin
	    try
	      let xhv2 = Hashtbl.find sigtmh x in
	      if xhv <> xhv2 then raise (Failure(x ^ " was explicitly assigned id " ^ xhv2 ^ " but this does not match its computed id " ^ xhv))
	    with Not_found -> ()
	  end;
	  Hashtbl.add sigtmh x xhv;
	  Hashtbl.add sigtmof xhv (i,agtp);
	  if i = 0 then
	    begin
	      try
		let itp = Hashtbl.find indextms xhv in
		if agtp <> itp then raise (Failure("The id " ^ xhv ^ " associated with the definition of " ^ x ^ " has is indexed to have the type " ^ tp_to_str itp ^ " not " ^ tp_to_str agtp ^ ". This indicates some fundamental bug since it would presumably imply a hash collision."));
	      with Not_found ->
		Hashtbl.add indextms xhv agtp
	    end;
	  let m = TmH xhv in
	  sigtm := (x,!aptmloc m)::!sigtm;
	  secstack := List.map (fun (y,f,atl,apl,st,sp) -> (y,f,atl,apl,(x,atl m)::st,sp)) !secstack;
	  if (!verbosity > 19) then (Printf.printf "i = %d\nbtm = %s\nbtp = %s\n" i (tm_to_str btm) (tp_to_str atp); flush stdout);
	end
  | AxDecl(x,a) ->
      if !reporteachitem then (Printf.printf "++ %s\n" x; flush stdout);
      let a = ltree_to_atree a in
      if List.mem_assoc x !sigtm || List.mem_assoc x !sigpf || List.mem x !ctxtp || List.mem_assoc x !ctxtm || List.mem_assoc x !ctxpf then
	raise (Failure(x ^ " has already been used."));
      let i = List.length !ctxtp in
      let atm = check_tm a Prop !polytm sigtmof !sigtm !ctxtp !ctxtm in
      let agtm = !tmallclos atm in
      let ahv = ptm_all_id (i,agtm) sigtmof sigdelta in
      add_sigdelta ahv (i,agtm);
      Hashtbl.add sigknh x ahv;
      begin
        if i = 0 then
          begin
            Hashtbl.add pfgknph ahv agtm;
          end;
	if !pfgout && i = 0 then pfgmain := PfgKnown(ahv,x,agtm)::!pfgmain
      end;
      activate_special_knowns ahv;
      begin
	if !sqlout then
	  begin
	    match !mainfilehash with
	    | Some docsha ->
		if !sqltermout then Printf.printf "INSERT INTO `term` (`termid`,`termtp`,`termpoly`) VALUES ('%s','%s',%d);\n" ahv (stp_html_string Prop) i;
		if not !presentationonly then (Printf.printf "INSERT INTO `termdoc` (`termid`,`docsha`,`termdocname`,`termdockind`) VALUES ('%s','%s',\"%s\",'a');\n" ahv docsha (String.escaped x));
	    | None -> ()
	  end
      end;
      sigpf := (x,!appfloc (Known(ahv)))::!sigpf;
      if i > 0 then (*** x will look polymorphic with i types after the appropriate section is ended ***)
	pushpolypf ((x,i),agtm);
      if not (Hashtbl.mem indexknowns ahv) &&
           begin
             if i = 0 then
               let pfgahv = pfg_propid agtm in
               not (Hashtbl.mem ownedprop pfgahv)
             else
               false
           end
      then
        begin
          Printf.printf "WARNING: The id %s for the proposition for axiom %s [pfg %s] is not indexed as previously known.\n" ahv x (Hash.hashval_hexstring (pfg_propid agtm));
        end;
      Hashtbl.add indexknowns ahv ();
      secstack := List.map (fun (y,f,atl,apl,st,sp) -> (y,f,atl,apl,st,(x,apl (Known(ahv)))::sp)) !secstack;
      if (!verbosity > 3) then (Printf.printf "Proposition of Axiom %s : %s was assigned id %s\n" x (tm_to_str agtm) ahv; flush stdout);
      ()
  | ThmDecl(c,x,a) ->
      let a = ltree_to_atree a in
      if List.mem_assoc x !sigtm || List.mem_assoc x !sigpf || List.mem x !ctxtp || List.mem_assoc x !ctxtm || List.mem_assoc x !ctxpf then
	raise (Failure(x ^ " has already been used."));
      if !includingsigfile then raise (Failure("Included signature file includes a theorem (" ^ x ^ "), but should only include axioms."));
      let i = List.length !ctxtp in
      let atm = check_tm a Prop !polytm sigtmof !sigtm !ctxtp !ctxtm in
      let agtm = !tmallclos atm in
      let ahv = ptm_all_id (i,agtm) sigtmof sigdelta in
      let pfgahv = pfg_propid agtm in
      Hashtbl.add sigknh x ahv;
      Hashtbl.add ownedprop pfgahv ();
      add_sigdelta ahv (i,agtm);
      if !reporteachitem then (Printf.printf "++ %s\nHASH %s\n" x ahv; flush stdout);
      if !sqlout then
	begin
	  match !treasure with
	  | Some(traddr) ->
	      begin
		match !mainfilehash with
		| Some docsha ->
		    if not !presentationonly then (Printf.printf "INSERT INTO `treasuretermdoc` (`treasureaddress`,`propid`,`docsha`,`thmdocname`) VALUES ('%s','%s','%s','%s');\n" traddr ahv docsha x);
		| None -> ()
	      end
	  | None -> ()
	end;
      if !verbosity > 5 && (Hashtbl.mem indexknowns ahv || Hashtbl.mem ownedprop pfgahv) then (Printf.printf "Warning: The id %s for the proposition for theorem %s is already known.\n" ahv x; flush stdout);
      if (!verbosity > 3) then (Printf.printf "Proposition of %s %s : %s was assigned id %s\n" c x (tm_to_str agtm) ahv; flush stdout);
      let currpflamclos = !pflamclos in
      deltaset := [];
      proving := Some (x,i,agtm,ahv,pfgahv);
      prooffun := (fun dl -> match dl with [d] -> currpflamclos d | _ -> raise (Failure("Bug with proof construction")));
      pfstate := [PfStateGoal(atm,!ctxtm,!ctxpf)];
      laststructaction := -1

let evaluate_docitem ditem =
  evaluate_docitem_1 ditem;
  match !latex with
  | Some hc ->
      begin
	if !thmsasexercises then
	  begin
	    match ditem with
	    | ThmDecl(_,x,_) ->
		begin
		  try
		    let h = Hashtbl.find sigknh x in
		    Printf.fprintf hc "\n$x%s\n" h
		  with Not_found -> ()
		end
	    | _ -> ()
	  end;
	output_docitem_latex hc ditem sigtmh sigknh;
	match ditem with
	| ThmDecl(_,x,_) ->
	    begin
	      match !inchan with
	      | Some(c) ->
		  begin
		    try
		      let h = Hashtbl.find sigknh x in
		      currtmid := h
		    with Not_found -> ()
		  end;
		  pflinestart := !lineno;
		  pfcharstart := !charno;
		  skip_to_line_char c inchanline inchanchar !lineno !charno;
	      | None -> ()
	    end
	| _ -> ()
      end
  | None -> ()

let rec structure_pfstate i goal1 pfstl =
  match pfstl with
  | (PfStateGoal(claim2,cxtm2,cxpf2)::pfstr) ->
      (PfStateSep(i,true)::goal1::structure_pfstate i (PfStateGoal(claim2,cxtm2,cxpf2)) pfstr)
  | pfstr ->
      (PfStateSep(i,true)::goal1::PfStateSep(i,false)::pfstr)

let print_pfstate () =
  List.iter
    (fun p ->
      match p with
      | PfStateSep(j,true) -> Printf.printf "PfStateSep(%d,true)\n" j
      | PfStateSep(j,false) -> Printf.printf "PfStateSep(%d,false)\n" j
      | PfStateGoal(claim1,_,_) -> Printf.printf "PfStateGoal(%s,_,_)\n" (tm_to_str claim1)
      )
    !pfstate

let evaluate_pftac_1 pitem thmname i gpgtm gphv pfggphv =
  match pitem with
  | PfStruct i when i < 4 ->
      if !verbosity > 19 then (Printf.printf "pfstruct %d\nLength of pfstate stack: %d\n" i (List.length !pfstate); print_pfstate (); flush stdout);
      begin
	match !pfstate with
	| PfStateSep(j,true)::pfstr ->
	    if i = j then
	      begin
		laststructaction := 2;
		pfstate := pfstr;
		if !verbosity > 19 then (Printf.printf "pfstruct %d pop\nLength of pfstate stack: %d\n" i (List.length !pfstate); flush stdout);
	      end
	    else
	      raise (Failure("Previous subproof not completed"))
	| PfStateSep(j,false)::pfstr ->
	    raise (Failure("Proof structuring bug"))
	| [] ->
	    raise (Failure("No claim to prove here"))
	| PfStateGoal(claim1,cxtm1,cxpf1)::PfStateGoal(claim2,cxtm2,cxpf2)::pfstr ->
	    laststructaction := 1;
	    pfstate := PfStateGoal(claim1,cxtm1,cxpf1)::structure_pfstate i (PfStateGoal(claim2,cxtm2,cxpf2)) pfstr;
	    if !verbosity > 19 then (Printf.printf "pfstruct %d push\nLength of pfstate stack: %d\n" i (List.length !pfstate); flush stdout);
	| PfStateGoal(claim1,cxtm1,cxpf1)::pfstr ->
	    raise (Failure("Inappropriate place for this proof structuring symbol since there is only one goal"))
      end
  | PfStruct 4 ->
      begin
	match !pfstate with
	| PfStateSep _::pfstr ->
	    raise (Failure("A subproof cannot be started here"))
	| [] ->
	    raise (Failure("No claim to prove here"))
	| PfStateGoal(claim1,cxtm1,cxpf1)::pfstr ->
	    pfstate := PfStateGoal(claim1,cxtm1,cxpf1)::PfStateSep(5,true)::pfstr
      end
  | PfStruct 5 ->
      begin
	match !pfstate with
	| PfStateSep(j,_)::pfstr ->
	    if j = 5 then
	      pfstate := pfstr
	    else
	      raise (Failure("Previous subproof not completed"))
	| [] ->
	    raise (Failure("Not proving a goal"))
	| _ -> raise (Failure("Subproof not completed"))
      end
  | Exact(d) ->
      laststructaction := 0;
      let d = ltree_to_atree d in
      begin
	match !pfstate with
	| PfStateGoal(claimtm,cxtm,cxpf)::pfstr ->
	    let dpf = check_pf d claimtm !polytm !polypf sigtmof sigdelta !sigtm !sigpf !ctxtp cxtm cxpf in
	    let currprooffun = !prooffun in
	    prooffun := (fun dl -> currprooffun (dpf::dl));
	    pfstate := pfstr
	| _ ->
	    raise (Failure("No current claim"))
      end
  | LetTac(xl,None) ->
      laststructaction := 0;
      begin
	match !pfstate with
	| PfStateGoal(claimtm,cxtm,cxpf)::pfstr ->
	    let rec let_tac_None_r xs claimtm cxtm cxpf =
	      match xs with
	      | [] ->
		  pfstate := PfStateGoal(claimtm,cxtm,cxpf)::pfstr
	      | (x::xr) ->
		  match claimtm with
		  | All(a,body) -> (*** If it's already an All, then don't call headnorm since headnorm will at least beta eta normalize and change the structure. ***)
		      let currprooffun = !prooffun in
		      prooffun := (fun dl -> match dl with d::dr -> currprooffun (TLam(a,d)::dr) | _ -> raise (Failure("proof reconstruction problem")));
		      let_tac_None_r xr body ((x,(a,None))::cxtm) (List.map (fun (y,q) -> (y,tmshift 0 1 q)) cxpf)
		  | _ -> (*** If it's not an All, then call headnorm and try to expose an All. ***)
		      let (p,dl) = headnorm claimtm sigdelta !deltaset in
		      match p with
		      | All(a,body) ->
			  let currprooffun = !prooffun in
			  prooffun := (fun dl -> match dl with d::dr -> currprooffun (TLam(a,d)::dr) | _ -> raise (Failure("proof reconstruction problem")));
			  let_tac_None_r xr body ((x,(a,None))::cxtm) (List.map (fun (y,q) -> (y,tmshift 0 1 q)) cxpf)
		      | _ ->
			  raise (Failure("let tactic used with " ^ x ^ " when claim is not a universal quantifier"))
	    in
	    let_tac_None_r xl claimtm cxtm cxpf
	| _ ->
	    raise (Failure("No current claim"))
      end
  | LetTac(xl,Some b) ->
      laststructaction := 0;
      let b = ltree_to_atree b in
      let btp = extract_tp b !ctxtp in
      begin
	match !pfstate with
	| PfStateGoal(claimtm,cxtm,cxpf)::pfstr ->
	    let rec let_tac_Some_r xs claimtm cxtm cxpf =
	      match xs with
	      | [] ->
		  pfstate := PfStateGoal(claimtm,cxtm,cxpf)::pfstr
	      | (x::xr) ->
		  match claimtm with
		  | All(a,body) -> (*** If it's already an All, then don't call headnorm since headnorm will at least beta eta normalize and change the structure. ***)
		      if a = btp then
			let currprooffun = !prooffun in
			prooffun := (fun dl -> match dl with d::dr -> currprooffun (TLam(a,d)::dr) | _ -> raise (Failure("proof reconstruction problem")));
			let_tac_Some_r xr body ((x,(a,None))::cxtm) (List.map (fun (y,q) -> (y,tmshift 0 1 q)) cxpf)
		      else
			raise (Failure(x ^ " ascribe type " ^ (tp_to_str btp) ^ " but the claim is universally quantified over type " ^ (tp_to_str a)))
		  | _ -> (*** If it's not an All, then call headnorm and try to expose an All. ***)
		      let (p,dl) = headnorm claimtm sigdelta !deltaset in
		      match p with
		      | All(a,body) ->
			  if a = btp then
			    let currprooffun = !prooffun in
			    prooffun := (fun dl -> match dl with d::dr -> currprooffun (TLam(a,d)::dr) | _ -> raise (Failure("proof reconstruction problem")));
			    let_tac_Some_r xr body ((x,(a,None))::cxtm) (List.map (fun (y,q) -> (y,tmshift 0 1 q)) cxpf)
			  else
			    raise (Failure(x ^ " ascribe type " ^ (tp_to_str btp) ^ " but the claim is universally quantified over type " ^ (tp_to_str a)))
		      | _ ->
			  raise (Failure("let tactic used with " ^ x ^ " when claim is not a universal quantifier"))
	    in
	    let_tac_Some_r xl claimtm cxtm cxpf
	| _ ->
	    raise (Failure("No current claim"))
      end
  | AssumeTac(xl,None) ->
      laststructaction := 0;
      begin
	match !pfstate with
	| PfStateGoal(claimtm,cxtm,cxpf)::pfstr ->
	    let rec assume_tac_None_r xs claimtm cxtm cxpf =
	      match xs with
	      | [] ->
		  pfstate := PfStateGoal(claimtm,cxtm,cxpf)::pfstr
	      | (x::xr) ->
		  match claimtm with
		  | Imp(p1,p2) -> (*** If it's already an Imp, then don't call headnorm since headnorm will at least beta eta normalize and change the structure. ***)
		      let currprooffun = !prooffun in
		      prooffun := (fun dl -> match dl with d::dr -> currprooffun (PLam(p1,d)::dr) | _ -> raise (Failure("proof reconstruction problem")));
		      assume_tac_None_r xr p2 cxtm ((x,p1)::cxpf)
		  | _ -> (*** If it's not an All, then call headnorm and try to expose an All. ***)
		      let (p,dl) = headnorm claimtm sigdelta !deltaset in
		      match p with
		      | Imp(p1,p2) ->
			  let currprooffun = !prooffun in
			  prooffun := (fun dl -> match dl with d::dr -> currprooffun (PLam(p1,d)::dr) | _ -> raise (Failure("proof reconstruction problem")));
			  assume_tac_None_r xr p2 cxtm ((x,p1)::cxpf)
		      | _ ->
			  raise (Failure("assume tactic used with " ^ x ^ " when claim is not an implication"))
	    in
	    assume_tac_None_r xl claimtm cxtm cxpf
	| _ ->
	    raise (Failure("No current claim"))
      end
  | AssumeTac(xl,Some b) ->
      laststructaction := 0;
      let b = ltree_to_atree b in
      begin
	match !pfstate with
	| PfStateGoal(claimtm,cxtm,cxpf)::pfstr ->
	    let btm = check_tm b Prop !polytm sigtmof !sigtm !ctxtp cxtm in
	    let rec assume_tac_Some_r xs claimtm cxtm cxpf =
	      match xs with
	      | [] ->
		  pfstate := PfStateGoal(claimtm,cxtm,cxpf)::pfstr
	      | (x::xr) ->
		  match claimtm with
		  | Imp(p1,p2) -> (*** If it's already an Imp, then don't call headnorm since headnorm will at least beta eta normalize and change the structure. ***)
		      begin
			match conv p1 btm sigdelta !deltaset with
			| Some(dl) ->
			    deltaset := dl;
			    let currprooffun = !prooffun in
			    prooffun := (fun dl -> match dl with d::dr -> currprooffun (PLam(btm,d)::dr) | _ -> raise (Failure("proof reconstruction problem")));
			    assume_tac_Some_r xr p2 cxtm ((x,btm)::cxpf)
			| None ->
			    raise (Failure(x ^ " ascribed prop " ^ (tm_to_str btm) ^ " but the antecendent of the claim is " ^ (tm_to_str p1)))
		      end
		  | _ -> (*** If it's not an Imp, then call headnorm and try to expose an Imp. ***)
		      let (p,dl) = headnorm claimtm sigdelta !deltaset in
		      match p with
		      | Imp(p1,p2) ->
			  begin
			    match conv p1 btm sigdelta !deltaset with
			    | Some(dl) ->
				deltaset := dl;
				let currprooffun = !prooffun in
				prooffun := (fun dl -> match dl with d::dr -> currprooffun (PLam(btm,d)::dr) | _ -> raise (Failure("proof reconstruction problem")));
				assume_tac_Some_r xr p2 cxtm ((x,btm)::cxpf)
			    | None ->
				raise (Failure(x ^ " ascribed prop " ^ (tm_to_str btm) ^ " but the antecendent of the claim is " ^ (tm_to_str p1)))
			  end
		      | _ ->
			  raise (Failure("assume tactic used with " ^ x ^ " when claim is not an implication"))
	    in
	    assume_tac_Some_r xl claimtm cxtm cxpf
	| _ ->
	    raise (Failure("No current claim"))
      end
  | SetTac(x,None,b) ->
      laststructaction := 0;
      let b = ltree_to_atree b in
      begin
	match !pfstate with
	| PfStateGoal(claimtm,cxtm,cxpf)::pfstr ->
	    let (btm,btp) = extract_tm b !polytm sigtmof !sigtm !ctxtp cxtm in
	    pfstate := (PfStateGoal(claimtm,(x,(btp,Some(btm)))::cxtm,cxpf))::pfstr
	| _ ->
	    raise (Failure("set tactic cannot be used when there is no claim"))
      end
  | SetTac(x,Some(a),b) ->
      let a = ltree_to_atree a in
      let b = ltree_to_atree b in
      begin
	match !pfstate with
	| PfStateGoal(claimtm,cxtm,cxpf)::pfstr ->
	    let atp = extract_tp a !ctxtp in
	    let btm = check_tm b atp !polytm sigtmof !sigtm !ctxtp cxtm in
	    pfstate := (PfStateGoal(claimtm,(x,(atp,Some(btm)))::cxtm,cxpf))::pfstr
	| _ ->
	    raise (Failure("set tactic cannot be used when there is no claim"))
      end
  | ApplyTac(a) ->
      laststructaction := 0;
      let a = ltree_to_atree a in
      if !verbosity > 19 then (Printf.printf "apply tactic begin\nLength of pfstate stack: %d\n" (List.length !pfstate); flush stdout);
      begin
	match !pfstate with
	| PfStateGoal(claimtm,cxtm,cxpf)::pfstr ->
	    let (apf,atm) = extract_pf a !polytm !polypf sigtmof sigdelta !sigtm !sigpf !ctxtp cxtm cxpf in
	    let sigma =
	      begin
		let i = ref 0 in
		List.iter (fun (_,(_,d)) -> match d with None -> incr i | _ -> ()) cxtm;
		map_for (fun j -> MDB j) 0 (!i-1)
	      end
	    in
	    let rec foapplyf p n margs subclaims apff =
	      try
		if !verbosity > 19 then (Printf.printf "about to call pattern_match with %s\n" (mtm_to_str p); flush stdout);
		let theta = pattern_match sigdelta p claimtm (fun _ -> raise Not_found) in
		if !verbosity > 19 then (Printf.printf "after pattern_match\n"; flush stdout);
		let ml = List.map (fun m -> mtm_to_tm (mtm_msub theta m)) margs in
		(*** The next code reverses the subclaims (so earliest arguments are to be proven first) and also grounds them from mtm to tm simultaneously ***)
		let subclaimtms = ref [] in
		List.iter
		  (fun c ->
		    let c1 = mtm_msub theta c in
		    let c2 = mtm_to_tm c1 in
		    subclaimtms := c2::!subclaimtms)
		  subclaims;
		(!subclaimtms,apff ml)
	      with MatchFail ->
		match p with
		| MImp(p1,p2) ->
		    foapplyf p2 n margs (p1::subclaims) (fun ml dl -> match dl with d::dr -> PPfAp(apff ml dr,d) | _ -> raise (Failure("proof reconstruction problem")))
		| MAll(a1,p2) ->
		    begin
		      match mtm_minap_db p2 0 with
		      | None -> (*** special case: no occurrence, use Eps _:a1, False ***)
			  let defelt = Ap(TpAp(Prim(0),a1),Lam(a1,All(Prop,DB(0)))) in
			  foapplyf p2 n margs subclaims (fun ml dl -> match dl with d::dr -> PTmAp(apff ml dr,defelt) | _ -> raise (Failure("proof reconstruction problem")))
		      | Some(l) when l = 0 -> (*** simple case, like a FO var ***)
			  foapplyf (mtm_ssub (MVar(n,sigma)::sigma) p2) (n+1) (MVar(n,sigma)::margs) subclaims (fun ml dl -> match ml with m::mr -> PTmAp(apff mr dl,m) | _ -> raise (Failure("proof reconstruction problem")))
		      | Some(l) -> (*** otherwise, move l arguments into the context of the metavar so that the higher-order pattern case is handled ***)
			  let sigmal = (map_for (fun j -> MDB j) 0 (l-1)) @ (List.map (mtm_shift 0 l) sigma) in
			  let rec lmvnr l a m =
			    if l > 0 then
			      begin
				match a with
				| Ar(a1,a2) -> MLam(a1,lmvnr (l-1) a2 m)
				| _ -> raise (Failure("Type Error found while attempting apply tactic"))
			      end
			    else
			      m
			  in
			  let lmvn = lmvnr l a1 (MVar(n,sigmal)) in
			  let p3 = mtm_ssub (lmvn::sigma) p2 in
			  let p4 = mtm_betared_if p3 (fun q _ -> mtm_lammvar_p q) in
			  foapplyf p4 (n+1) (lmvn::margs) subclaims (fun ml dl -> match ml with m::mr -> PTmAp(apff mr dl,m) | _ -> raise (Failure("proof reconstruction problem")))
		    end
		| _ ->
		    begin
		      let p1 = mtm_betared_if p (fun _ _ -> true) in (*** try to beta reduce ***)
		      if p = p1 then (*** there were no beta reductions ***)
			let (p2,del) = mheadnorm p sigdelta !deltaset in (*** try to delta expand some heads ***)
			if p = p2 then (*** no delta expansions ***)
			  raise Not_found (*** give up and fail ***)
			else
			  begin
			    deltaset := del;
			    foapplyf p2 n margs subclaims apff
			  end
		      else
			foapplyf p1 n margs subclaims apff
		    end
	    in
	    begin
	      try
		if !verbosity > 19 then (Printf.printf "Proof term given with apply proves %s\n" (tm_to_str atm); flush stdout);
		let (subclaims,apff) = foapplyf (tm_to_mtm atm) 0 [] [] (fun ml dl -> apf) in
		let nsubs = List.length subclaims in
		let currprooffun = !prooffun in
		prooffun := (fun dl -> let (dl1,dl2) = split_list nsubs dl in currprooffun (apff (List.rev dl1)::dl2));
		pfstate := (List.map (fun c -> PfStateGoal(c,cxtm,cxpf)) subclaims) @ pfstr;
		if !verbosity > 19 then (Printf.printf "here nach apply\nLength of pfstate stack: %d\n" (List.length !pfstate); flush stdout);
	      with Not_found ->
		raise (Failure("apply does not match the current claim"))
	    end
	| _ ->
	    raise (Failure("apply tactic cannot be used when there is no claim"))
      end
  | ClaimTac(x,a) ->
      laststructaction := 0;
      let a = ltree_to_atree a in
      begin
	match !pfstate with
	| PfStateGoal(claimtm,cxtm,cxpf)::pfstr ->
	    let atm = check_tm a Prop !polytm sigtmof !sigtm !ctxtp cxtm in
	    pfstate := (PfStateGoal(atm,cxtm,cxpf))::(PfStateGoal(claimtm,cxtm,(x,atm)::cxpf))::pfstr;
	    let currprooffun = !prooffun in
	    prooffun := (fun dl -> match dl with d1::d2::dr -> currprooffun (PPfAp(PLam(atm,d2),d1)::dr) | _ -> raise (Failure("proof reconstruction problem")))
	| _ ->
	    raise (Failure("claim tactic cannot be used when there is no claim"))
      end
  | ProveTac(a,[]) ->
      laststructaction := 0;
      let a = ltree_to_atree a in
      begin
	match !pfstate with
	| PfStateGoal(claimtm,cxtm,cxpf)::pfstr ->
	    let atm = check_tm a Prop !polytm sigtmof !sigtm !ctxtp cxtm in
	    begin
	      match conv claimtm atm sigdelta !deltaset with
	      | Some(dl) ->
		  deltaset := dl;
		  pfstate := (PfStateGoal(atm,cxtm,cxpf))::pfstr
	      | None ->
		  match conv (TmH(!fal)) atm sigdelta !deltaset with (*** or if prove False, then use FalseE ***)
                  | Some(del) ->
		     deltaset := del;
		     let currprooffun = !prooffun in
		     prooffun := (fun dl -> match dl with d1::dr -> currprooffun (PTmAp(PPfAp(Known(!fale),d1),claimtm)::dr) | _ -> raise (Failure("proof reconstruction problem")));
		     pfstate := (PfStateGoal(atm,cxtm,cxpf))::pfstr
                  | None ->
		     match conv (All(Prop,DB(0))) atm sigdelta !deltaset with (*** or if prove False, then use old FalseE ***)
		     | Some(del) ->
		        deltaset := del;
		        let currprooffun = !prooffun in
		        prooffun := (fun dl -> match dl with d1::dr -> currprooffun (PTmAp(d1,claimtm)::dr) | _ -> raise (Failure("proof reconstruction problem")));
		        pfstate := (PfStateGoal(atm,cxtm,cxpf))::pfstr
		     | None ->
		        raise (Failure("Proposition given with prove tactic does not match the current claim.\n" ^ tm_to_str claimtm ^ "\n" ^ tm_to_str atm))
	    end
	| _ ->
	    raise (Failure("prove tactic cannot be used when there is no claim"))
      end
  | ProveTac(a,bl) ->
      laststructaction := 0;
      raise (Failure("prove tactic can currently only be used with one proposition"))
  | CasesTac(a,xbll) ->
      laststructaction := 0;
      raise (Failure("cases tactic is not yet implemented"))
  | WitnessTac(a) ->
      laststructaction := 0;
      let a = ltree_to_atree a in
      begin
	match !pfstate with
	| PfStateGoal(claimtm,cxtm,cxpf)::pfstr ->
	   begin
	     try
	       let (etp,ep) = extract_exclaim claimtm in
	       let atm = check_tm a etp !polytm sigtmof !sigtm !ctxtp cxtm in
	       let epatm =
		 begin
		   match ep with
		   | Lam(_,epbody) -> tmsubst epbody 0 atm
		   | _ -> Ap(ep,atm)
		 end
	       in
	       if !verbosity > 50 then (Printf.printf "witness tactic with etp = %s\n and ep = %s\n and atm = %s\n and epatm = %s\n" (tp_to_str etp) (tm_to_str ep) (tm_to_str atm) (tm_to_str epatm); flush stdout);
	       pfstate := (PfStateGoal(epatm,cxtm,cxpf)::pfstr);
	       let currprooffun = !prooffun in
               if !expolyIknown then
	         prooffun := (fun dl -> match dl with d1::dr -> currprooffun (PPfAp(PTmAp(PTmAp(PTpAp(Known(!expolyI),etp),ep),atm),d1)::dr) | _ -> raise (Failure("proof reconstruction problem")))
               else
	         prooffun :=
                   (fun dl ->
                     match dl with
                       d1::dr ->
                       currprooffun (TLam(Prop,PLam(All(etp,Imp(gen_lam_body(tmshift 0 1 ep),DB(1))),PPfAp(PTmAp(Hyp(0),tmshift 0 1 atm),pftmshift 0 1 (pfshift 0 1 d1))))::dr)
                     | _ -> raise (Failure("proof reconstruction problem")))
	     with Not_found ->
	       raise (Failure("witness tactic can only be used when claim is existential"))
	   end
	| _ ->
	   raise (Failure("witness tactic cannot be used when there is no claim"))
      end
  | RewriteTac(sym,a,posl) -> (*** rewrite from right to left if not sym or left to right if sym ***)
      laststructaction := 0;
      let a = ltree_to_atree a in
      if !verbosity > 19 then (Printf.printf "rewrite tactic begin\nLength of pfstate stack: %d\n" (List.length !pfstate); flush stdout);
      begin
	match !pfstate with
	| PfStateGoal(claimtm,cxtm,cxpf)::pfstr ->
	    let (apf,atm) = extract_pf a !polytm !polypf sigtmof sigdelta !sigtm !sigpf !ctxtp cxtm cxpf in
	    let sigma =
	      begin
		let i = ref 0 in
		List.iter (fun (_,(_,d)) -> match d with None -> incr i | _ -> ()) cxtm;
		map_for (fun j -> MDB j) 0 (!i-1)
	      end
	    in
	    let inpos i = match posl with [] -> true | _ -> List.mem i posl in
	    let posr = ref 0 in
	    let rec forewritef4 z etmi mtm =
	      if !verbosity > 79 then (Printf.printf "forewritef4: %d\n etmi: %s\n mtm: %s\n" z (tm_to_str etmi) (tm_to_str mtm); flush stdout);
	      if mtm = etmi then
		begin
		  incr posr;
		  if inpos !posr then
		    DB(z)
		  else
		    forewritef5 z etmi mtm
		end
	      else
		forewritef5 z etmi mtm
	    and forewritef5 z etmi mtm =
	      if !verbosity > 79 then (Printf.printf "forewritef5: %d\n etmi: %s\n mtm: %s\n" z (tm_to_str etmi) (tm_to_str mtm); flush stdout);
	      match mtm with
	      | Ap(m1,m2) ->
		  let m1b = forewritef4 z etmi m1 in
		  let m2b = forewritef4 z etmi m2 in
		  Ap(m1b,m2b)
	      | Imp(m1,m2) ->
		  let m1b = forewritef4 z etmi m1 in
		  let m2b = forewritef4 z etmi m2 in
		  Imp(m1b,m2b)
	      | Lam(a1,m1) -> Lam(a1,forewritef4 (z+1) (tmshift 0 1 etmi) m1)
	      | All(a1,m1) -> All(a1,forewritef4 (z+1) (tmshift 0 1 etmi) m1)
	      | DB(j) when j >= z -> DB(j+1)
	      | _ -> mtm
	    in
	    let rec forewritef2 z n etp etm mtm =
	      if !verbosity > 79 then (Printf.printf "forewritef2: %d %d\n etm: %s\n mtm: %s\n" z n (mtm_to_str etm) (tm_to_str mtm); flush stdout);
	      begin
		try
		  begin
		    if !verbosity > 19 then (Printf.printf "rewrite: about to call pattern_match with\n    %s\n =? %s\n" (mtm_to_str etm) (tm_to_str mtm); flush stdout);
		    let theta = pattern_match sigdelta etm mtm (fun _ -> raise Not_found) in
		    if !verbosity > 19 then (Printf.printf "after pattern_match %d\n" !posr; flush stdout);
		    try
		      begin
			for i = 0 to n-1 do
			  let _ = theta i in ()
			done;
			incr posr;
			if inpos !posr then
			  begin
			    (*** the instantiation to use ***)
			    (DB(z),theta)
			  end
			else
			  forewritef3 z n etp etm mtm
		      end
		    with Not_found ->
		      forewritef3 z n etp etm mtm
		  end
		with MatchFail ->
		  forewritef3 z n etp etm mtm
	      end
	    and forewritef3 z n etp etm mtm =
	      if !verbosity > 79 then (Printf.printf "forewritef3: %d %d\n etm: %s\n mtm: %s\n" z n (mtm_to_str etm) (tm_to_str mtm); flush stdout);
	      match mtm with
	      | Ap(m1,m2) ->
		  begin
		    try
		      let (leibp1,theta) = forewritef2 z n etp etm m1 in
		      let leibp2 = forewritef4 z (mtm_to_tm (mtm_msub theta etm)) m2 in
		      (Ap(leibp1,leibp2),theta)
		    with Not_found ->
		      let (leibp2,theta) = forewritef2 z n etp etm m2 in
		      (Ap(tmshift z 1 m1,leibp2),theta)
		  end
	      | Imp(m1,m2) ->
		  begin
		    try
		      let (leibp1,theta) = forewritef2 z n etp etm m1 in
		      let leibp2 = forewritef4 z (mtm_to_tm (mtm_msub theta etm)) m2 in
		      (Imp(leibp1,leibp2),theta)
		    with Not_found ->
		      let (leibp2,theta) = forewritef2 z n etp etm m2 in
		      (Imp(tmshift z 1 m1,leibp2),theta)
		  end
	      | Lam(a1,m1) ->
		  let (leibp1,theta) = forewritef2 (z+1) n etp (mtm_shift 0 1 etm) m1 in
		  (Lam(a1,leibp1),theta)
	      | All(a1,m1) ->
		  let (leibp1,theta) = forewritef2 (z+1) n etp (mtm_shift 0 1 etm) m1 in
		  (All(a1,leibp1),theta)
	      | _ -> raise Not_found
	    in
	    let destruct_leibeq p =
	      match p with
	      | MAp(MAp(MTpAp(MTmH(e),etp),ltm),rtm) when e = !eqPoly -> (etp,ltm,rtm) (*** recognize it before delta expanding ***)
	      | MAll(Ar(etp,Prop),MImp(MAp(MDB(0),shltm),MAp(MDB(0),shrtm))) -> (*** also need to recognize delta expanded version to avoid confusion ***)
		  begin
		    try
		      let ltm = mtm_shift 0 (-1) shltm in
		      let rtm = mtm_shift 0 (-1) shrtm in
		      (etp,ltm,rtm)
		    with NegDB -> raise Not_found
		  end
	      | _ ->
		  raise Not_found
	    in
	    let rec forewritef p n margs subclaims apff =
	      if !verbosity > 19 then (Printf.printf "forewritef p : %s\n" (mtm_to_str p); flush stdout);
	      begin
		try
		  let (etp,ltm,rtm) = destruct_leibeq p in
		  begin
		    try
		      let tm1 = if sym then rtm else ltm in
		      let tm2 = if sym then ltm else rtm in
		      let (leibq,theta) = forewritef2 0 n etp tm1 claimtm in
		      if !verbosity > 90 then (Printf.printf "leibq: %s\n" (tm_to_str leibq); flush stdout);
		      let tm2i = mtm_to_tm (mtm_msub theta tm2) in
		      let ml = List.map (fun m -> mtm_to_tm (mtm_msub theta m)) margs in
		      let subclaimtms = ref [] in
		      List.iter
			(fun c ->
			  let c1 = mtm_msub theta c in
			  let c2 = mtm_to_tm c1 in
			  subclaimtms := c2::!subclaimtms)
			subclaims;
		      let apffm =
			if sym then
			  (fun dl ->
			    match dl with
			      d::dr ->
				let epf = apff ml dr in
				PPfAp(PTmAp(epf,Lam(etp,Lam(etp,tmshift 0 1 leibq))),d)
			    | _ -> raise (Failure("proof reconstruction problem")))
			else
			  (fun dl ->
			    match dl with
			      d::dr ->
				let epf = apff ml dr in
				PPfAp(PTmAp(epf,Lam(etp,Lam(etp,tmshift 1 1 leibq))),d)
			    | _ -> raise (Failure("proof reconstruction problem")))
		      in
		      (tmsubst leibq 0 tm2i::!subclaimtms,apffm)
		    with Not_found ->
		      raise (Failure("rewrite tactic failed"))
		  end
		with Not_found ->
		  match p with
		  | MImp(p1,p2) ->
		      forewritef p2 n margs (p1::subclaims) (fun ml dl -> match dl with d::dr -> PPfAp(apff ml dr,d) | _ -> raise (Failure("proof reconstruction problem")))
		  | MAll(a1,p2) ->
		      begin
			match mtm_minap_db p2 0 with
			| None -> (*** special case: no occurrence, use Eps _:a1, False ***)
			    let defelt = Ap(TpAp(Prim(0),a1),Lam(a1,All(Prop,DB(0)))) in
			    forewritef p2 n margs subclaims (fun ml dl -> match dl with d::dr -> PTmAp(apff ml dr,defelt) | _ -> raise (Failure("proof reconstruction problem")))
			| Some(l) when l = 0 -> (*** simple case, like a FO var ***)
			    forewritef (mtm_ssub (MVar(n,sigma)::sigma) p2) (n+1) (MVar(n,sigma)::margs) subclaims (fun ml dl -> match ml with m::mr -> PTmAp(apff mr dl,m) | _ -> raise (Failure("proof reconstruction problem")))
			| Some(l) -> (*** otherwise, move l arguments into the context of the metavar so that the higher-order pattern case is handled ***)
			    let sigmal = (map_for (fun j -> MDB j) 0 (l-1)) @ (List.map (mtm_shift 0 l) sigma) in
			    let rec lmvnr l a m =
			      if l > 0 then
				begin
				  match a with
				  | Ar(a1,a2) -> MLam(a1,lmvnr (l-1) a2 m)
				  | _ -> raise (Failure("Type Error found while attempting apply tactic"))
				end
			      else
				m
			    in
			    let lmvn = lmvnr l a1 (MVar(n,sigmal)) in
			    let p3 = mtm_ssub (lmvn::sigma) p2 in
			    let p4 = mtm_betared_if p3 (fun q _ -> mtm_lammvar_p q) in
			    forewritef p4 (n+1) (lmvn::margs) subclaims (fun ml dl -> match ml with m::mr -> PTmAp(apff mr dl,m) | _ -> raise (Failure("proof reconstruction problem")))
		      end
		  | _ ->
		      begin
			let p1 = mtm_betared_if p (fun _ _ -> true) in (*** try to beta reduce ***)
			if p = p1 then (*** there were no beta reductions ***)
			  let (p2,del) = mheadnorm p sigdelta !deltaset in (*** try to delta expand some heads ***)
			  if p = p2 then
			    begin
			      if !verbosity > 19 then (Printf.printf "p : %s\n" (mtm_to_str p); flush stdout);
			      raise (Failure("rewrite tactic given a proof of a non-equation"))
			    end
			  else
			    begin
			      deltaset := del;
			      forewritef p2 n margs subclaims apff
			    end
			else
			  forewritef p1 n margs subclaims apff
		      end
	      end
	    in
	    begin
	      try
		if !verbosity > 19 then (Printf.printf "Proof term given with rewrite proves %s\n" (tm_to_str atm); flush stdout);
		let (subclaims,apff) = forewritef (tm_to_mtm atm) 0 [] [] (fun ml dl -> apf) in
		let nsubs = List.length subclaims in
		let currprooffun = !prooffun in
		prooffun := (fun dl -> match dl with d::dr -> let (dl1,dl2) = split_list (nsubs-1) dr in currprooffun (apff (d::(List.rev dl1))::dl2) | [] -> raise (Failure("proof reconstruction problem")));
		pfstate := (List.map (fun c -> PfStateGoal(c,cxtm,cxpf)) subclaims) @ pfstr;
		if !verbosity > 19 then (Printf.printf "here nach apply\nLength of pfstate stack: %d\n" (List.length !pfstate); flush stdout);
	      with Not_found ->
		raise (Failure("rewrite tactic failed"))
	    end
	| _ ->
	    raise (Failure("rewrite tactic cannot be used when there is no claim"))
      end
  | Qed ->
      begin
	if !pfstate = [] then
	  begin
	    try
	      if !verbosity > 19 then (Printf.printf "Qed start\n"; flush stdout);
	      if not (Hashtbl.mem indexknowns gphv) then Hashtbl.add indexknowns gphv ();
	      if not (Hashtbl.mem ownedprop pfggphv) then Hashtbl.add ownedprop pfggphv ();
	      activate_special_knowns gphv;
	      sigpf := (thmname,!appfloc (Known(gphv)))::!sigpf;
	      if i > 0 then (*** x will look polymorphic with i types after the appropriate section is ended ***)
		pushpolypf ((thmname,i),gpgtm);
	      secstack := List.map (fun (y,f,atl,apl,st,sp) -> (y,f,atl,apl,st,(thmname,apl (Known(gphv)))::sp)) !secstack;
	      proving := None;
	      let dgpf = !prooffun [] in
	      begin
                if i = 0 then
                  begin
                    Hashtbl.add pfgknph gphv gpgtm;
                  end;
	        if !pfgout && i = 0 then
	          pfgmain := PfgThm(gphv,thmname,gpgtm,dgpf)::!pfgmain
	      end;
	      if (!verbosity > 19) then (Printf.printf "Double checking:\n%s\n%s\n" (pf_to_str dgpf) (tm_to_str gpgtm); flush stdout);
	      match check_propofpf sigdelta sigtmof [] [] dgpf gpgtm !deltaset with
	      | None ->
		  if (!verbosity > 19) then (Printf.printf "Proof doesn't doublecheck!\n%s\n" (pf_to_str dgpf); flush stdout);
		  raise (Failure("Proof doesn't prove the proposition."))
	      | Some(dl) ->
		  deltaset := dl;
                  if !pfgout then List.iter (fun d -> Hashtbl.add pfgdelta d ()) !deltaset;
		  if (!verbosity > 19) then (Printf.printf "Delta Set:"; List.iter (fun h -> Printf.printf " %s" h) dl; Printf.printf "\n"; flush stdout);
		  let dhv = ppf_id (i,dgpf) sigtmof sigdelta in
		  if (!verbosity > 3) then (Printf.printf "Proof %s\n of %s was assigned id %s\n" (pf_to_str dgpf) thmname dhv; flush stdout);
		  if (!reportpfcomplexity) then (Printf.printf "(PFCOMPLEXITY \"%s\" \"%s\" \"%s\" %d)\n" thmname gphv dhv (pf_complexity dgpf); flush stdout);
		  begin
		    if !sqlout then
		      begin
			match !mainfilehash with
			| Some docsha ->
			    if !sqltermout then Printf.printf "INSERT INTO `term` (`termid`,`termtp`,`termpoly`) VALUES ('%s','%s',%d);\n" gphv (stp_html_string Prop) i;
			    if not !presentationonly then (Printf.printf "INSERT INTO `termdoc` (`termid`,`docsha`,`termdocname`,`termdockind`) VALUES ('%s','%s',\"%s\",'T');\n" gphv docsha (String.escaped thmname));
			    if !sqltermout then Printf.printf "INSERT INTO `proppf` (`propid`,`pfid`) VALUES ('%s','%s');\n" gphv dhv;
			    if not !presentationonly then
			      begin
				Printf.printf "INSERT INTO `proppfdoc` (`propid`,`pfid`,`docsha`) VALUES ('%s','%s',\"%s\");\n" gphv dhv docsha;
				List.iter
				  (fun d ->
				    Printf.printf "INSERT INTO `proppfdocdelta` (`propid`,`pfid`,`docsha`,`termid`) VALUES ('%s','%s','%s','%s');\n" gphv dhv docsha d)
				  !deltaset
			      end
			| None -> ()
		      end
		  end;
	    with AdmittedPf ->
	      if (!verbosity > 9) then (Printf.printf "Theorem %s admitted\n" thmname; flush stdout);
              if !pfgout && i = 0 then pfgmain := PfgConj(gphv,thmname,gpgtm)::!pfgmain;
	      if (!ajax && !ajaxactive) then (Printf.printf "I$"; exit 1);
	      begin
		if !sqlout then
		  begin
		    match !mainfilehash with
		    | Some docsha ->
			if !sqltermout then Printf.printf "INSERT INTO `term` (`termid`,`termtp`,`termpoly`) VALUES ('%s','%s',%d);\n" gphv (stp_html_string Prop) i;
			if not !presentationonly then (Printf.printf "INSERT INTO `termdoc` (`termid`,`docsha`,`termdocname`,`termdockind`) VALUES ('%s','%s',\"%s\",'t');\n" gphv docsha (String.escaped thmname));
		    | None -> ()
		  end
	      end;
	      treasure := None;
	  end
	else
	  raise (Failure("Proof of " ^ thmname ^ " is incomplete"))
      end
  | Admitted ->
      begin
        if !pfgout && i = 0 then pfgmain := PfgConj(gphv,thmname,gpgtm)::!pfgmain;
	if i > 0 then (*** x will look polymorphic with i types after the appropriate section is ended ***)
	  pushpolypf ((thmname,i),gpgtm);
	activate_special_knowns gphv;
	sigpf := (thmname,!appfloc (Known(gphv)))::!sigpf;
	secstack := List.map (fun (y,f,atl,apl,st,sp) -> (y,f,atl,apl,st,(thmname,apl (Known(gphv)))::sp)) !secstack;
	proving := None;
	pfstate := [];
	treasure := None;
	admits := true;
	begin
	  if !sqlout then
	    begin
	      match !mainfilehash with
	      | Some docsha ->
		  if !sqltermout then Printf.printf "INSERT INTO `term` (`termid`,`termtp`,`termpoly`) VALUES ('%s','%s',%d);\n" gphv (stp_html_string Prop) i;
		  if not !presentationonly then (Printf.printf "INSERT INTO `termdoc` (`termid`,`docsha`,`termdocname`,`termdockind`) VALUES ('%s','%s',\"%s\",'t');\n" gphv docsha (String.escaped thmname));
	      | None -> ()
	    end
	end;
      end
  | Admit ->
      begin
	match !pfstate with
	| (pfst::pfstr) ->
	    pfstate := pfstr;
	    admits := true;
	    prooffun := (fun _ -> raise AdmittedPf)
	| [] -> raise (Failure("No goal to admit"))
      end
  | _ ->
      raise (Failure("Unknown proof tactic"))

let rec evaluate_pftac_2 () =
  match !pfstate with
  | PfStateSep(j,false)::pfstr ->
      begin
	match !latex with
	| Some hc -> output_pftacitem_latex hc (PfStruct(j)) sigtmh sigknh 3
	| None -> ()
      end;
      pfstate := pfstr;
      evaluate_pftac_2 ()
  | _ -> ()

let evaluate_pftac pitem thmname i gpgtm gphv pfggphv =
  if !verbosity > 19 then (Printf.printf "pre1 pfstruct %d\nLength of pfstate stack: %d\n" i (List.length !pfstate); print_pfstate (); flush stdout);
  evaluate_pftac_1 pitem thmname i gpgtm gphv pfggphv;
  begin
    match !latex with
    | Some hc ->
(*	if pitem = Qed || pitem = Admitted then
	  begin
	    match !inchan with
	    | Some(c) -> buffer_to_line_char c pftext inchanline inchanchar !lineno !charno
	    | None -> ()
	  end; *)
	output_pftacitem_latex hc pitem sigtmh sigknh !laststructaction
    | None -> ()
  end;
  if !verbosity > 19 then (Printf.printf "pre2 pfstruct %d\nLength of pfstate stack: %d\n" i (List.length !pfstate); print_pfstate (); flush stdout);
  evaluate_pftac_2 ()

let init_env () =
  ctxtp := [];
  ctxtm := [];
  ctxpf := [];
  tparclos := (fun a -> a);
  tmallclos := (fun m -> m);
  tmlamclos := (fun m -> m);
  pflamclos := (fun d -> d);
  aptmloc := (fun m cxtp cxtm -> m);
  appfloc := (fun d cxtp cxtm cxpf -> d);
  secstack := [];
  popfn := (fun () -> ())
	
(*** Function for checking if a file solves a problem file in addition to checking the solution file for correctness. ***)
let mgchecksolves probc solnc =
  init_env ();
  let ptl = ref (TokStrRest(Lexer.token,Lexing.from_channel probc)) in
  let stl = ref (TokStrRest(Lexer.token,Lexing.from_channel solnc)) in
  let currentthm = ref "" in
  let inadmitted = ref false in
  let linenop = ref 1 in
  let charnop = ref 0 in
  let linenos = ref 1 in
  let charnos = ref 0 in
  try
    while true do
      match !proving with
      | None ->
	  begin
	    lineno := !linenop;
	    charno := !charnop;
	    let (pditem,ptr) =
	      try
		parse_docitem !ptl
	      with Lexer.Eof ->
		let _ = parse_docitem !stl in
		raise (Failure("Problem file ended prematurely"))
	    in
	    linenop := !lineno;
	    charnop := !charno;
	    lineno := !linenos;
	    charno := !charnos;
	    authors := [];
	    title := None;
	    let (sditem,str) =
	      try
		parse_docitem !stl
	      with Lexer.Eof ->
		raise (Failure("Solution file ended prematurely"))
	    in
	    linenos := !lineno;
	    charnos := !charno;
	    ptl := ptr;
	    stl := str;
	    evaluate_docitem sditem;
	    if pditem <> sditem then
	      raise (Failure("Solution document differs from problem document at line " ^ (string_of_int !linenop) ^ " and char " ^ (string_of_int !charnop) ^ " / line " ^ (string_of_int !linenos) ^ " and char " ^ (string_of_int !charnos)));
	    match sditem with
	    | ThmDecl(_,x,_) -> currentthm := x
	    | _ -> ()
	  end
      | Some (thmname,i,gpgtm,gphv,pfggphv) -> (*** reading a proof in the solution file ***)
	  begin
	    if !inadmitted then
	      begin
		lineno := !linenos;
		charno := !charnos;
		let (spftac,str) =
		  try
		    parse_pftacitem !stl
		  with Lexer.Eof ->
		    raise (Failure("Solution file ended prematurely"))
		in
		stl := str;
		linenos := !lineno;
		charnos := !charno;
		evaluate_pftac spftac thmname i gpgtm gphv pfggphv;
		if spftac = Qed then
		  inadmitted := false
		else if spftac = Admit || spftac = Admitted then
		  raise (Failure("Incomplete proof in solution file"))
	      end
	    else
	      begin
		lineno := !linenop;
		charno := !charnop;
		let (ppftac,ptr) =
		  try
		    parse_pftacitem !ptl
		  with Lexer.Eof ->
		    let _ = parse_pftacitem !stl in
		    raise (Failure("Problem file ended prematurely"))
		in
		linenop := !lineno;
		charnop := !charno;
		lineno := !linenos;
		charno := !charnos;
		let (spftac,str) =
		  try
		    parse_pftacitem !stl
		  with Lexer.Eof ->
		    raise (Failure("Solution file ended prematurely"))
		in
		linenos := !lineno;
		charnos := !charno;
		ptl := ptr;
		stl := str;
		evaluate_pftac spftac thmname i gpgtm gphv pfggphv;
		if spftac = Admit || spftac = Admitted then
		  raise (Failure("Incomplete proof in solution file"));
		if ppftac <> spftac then
		  begin
		    if ppftac = Admitted then
		      begin
			exercises := !currentthm :: !exercises;
			if spftac <> Qed then
			  inadmitted := true
		      end
		    else
		      raise (Failure("Solution document differs from problem document in proof at line " ^ (string_of_int !linenop) ^ " and char " ^ (string_of_int !charnop) ^ " / line " ^ (string_of_int !linenos) ^ " and char " ^ (string_of_int !charnos)))
		  end
	      end
	  end
    done
  with
  | Lexer.Eof ->
      ()
  | ParsingError(x,l1,c1,l2,c2) ->
      if !webout then
	begin
          Printf.printf "AS%d:%d:%d:%d\n"  l1 c1 l2 c2;
	  Printf.printf "<div class='documentfail'>Syntax error between line %d char %d and line %d char %d:<br/>%s</div>" l1 c1 l2 c2 x;
	  exit 1
	end
      else
	begin
	  Printf.printf "Syntax error between line %d char %d and line %d char %d:\n%s\n" l1 c1 l2 c2 x;
	  exit 1
	end
  | Failure(x) ->
      if !webout then
	begin
          Printf.printf "AF%d:%d\n"  !lineno !charno;
	  Printf.printf "<div class='documentfail'>Failure at line %d char %d: %s</div>" !lineno !charno x; flush stdout;
	  exit 1
	end
      else
	begin
	  Printf.printf "Failure at line %d char %d: %s\n" !lineno !charno x; flush stdout;
	  exit 1
	end

let mgcheck_ajax c =
  let tl = ref (TokStrRest(Lexer.token,Lexing.from_channel c)) in
  lineno := 1;
  charno := 0;
  verbosity := 1;
  try
    while true do
      if !verbosity > 59 then (Printf.printf "Main Loop Start\n"; flush stdout);
      match !proving with
      | None ->
	  ignore (parse_docitem !tl); (*** hopefully this raises End_of_file ***)
	  raise (Failure("The content must not continue beyond the proof."))
      | Some (thmname,i,gpgtm,gphv,pfggphv) -> (*** reading a proof ***)
	  let (pitem,tr) = parse_pftacitem !tl in
	  tl := tr;
	  evaluate_pftac pitem thmname i gpgtm gphv pfggphv
    done
  with
  | Lexer.Eof ->
      begin
	match !proving with
	| None -> () (* OK *)
	| Some(thmname,i,gpgtm,gphv,pfggphv) -> (* pretend there is a Qed at the end *)
	    try
	      evaluate_pftac Qed thmname i gpgtm gphv pfggphv
	    with
	    | Failure(x) ->
		Printf.printf "F %d %d$%s" !lineno !charno x
	    | ParsingError(x,l1,c1,l2,c2) ->
		Printf.printf "E %d %d %d %d$%s" l1 c1 l2 c2 x
      end
  | Failure(x) ->
      Printf.printf "F %d %d$%s" !lineno !charno x
  | ParsingError(x,l1,c1,l2,c2) ->
      Printf.printf "E %d %d %d %d$%s" l1 c1 l2 c2 x

(*** Main function for checking a file ***)
let mgcheck c =
  init_env ();
  let tl = ref (TokStrRest(Lexer.token,Lexing.from_channel c)) in
  lineno := 1;
  charno := 0;
  try
    while true do
      if !verbosity > 59 then (Printf.printf "Main Loop Start\n"; flush stdout);
      match !proving with
      | None ->
	  let (ditem,tr) = parse_docitem !tl in
	  tl := tr;
	  evaluate_docitem ditem
      | Some (thmname,i,gpgtm,gphv,pfggphv) -> (*** reading a proof ***)
	  let (pitem,tr) = parse_pftacitem !tl in
	  tl := tr;
	  evaluate_pftac pitem thmname i gpgtm gphv pfggphv
    done
  with
  | Lexer.Eof ->
    begin
      if !ajax then (*** we should be expecting a proof which can be read from the given proof file ***)
	if not !includingsigfile then
	  begin
	    ajaxactive := true;
	    let ca = open_in !ajaxpffile in
	    mgcheck_ajax ca
	  end
      else
	if (!verbosity > 9) then (Printf.printf "done.\n"; flush stdout)
    end
  | ParsingError(x,l1,c1,l2,c2) ->
      if !webout then
	begin
          Printf.printf "AS%d:%d:%d:%d\n"  l1 c1 l2 c2;
	  Printf.printf "<div class='documentfail'>Syntax error between line %d char %d and line %d char %d:<br/>%s</div>" l1 c1 l2 c2 x;
	  exit 1
	end
      else if !ajax then
	begin
	  Printf.printf "f\n"; (*** this indicates a fundamental problem, not a problem with the ajax input ***)
	  exit 1
	end
      else
	begin
	  Printf.printf "Syntax error between line %d char %d and line %d char %d:\n%s\n" l1 c1 l2 c2 x;
	  exit 1
	end
  | Failure(x) ->
      if !webout then
	begin
          Printf.printf "AF%d:%d\n"  !lineno !charno;
	  Printf.printf "<div class='documentfail'>Failure at line %d char %d: %s</div>" !lineno !charno x; flush stdout;
	  exit 1
	end
      else if !ajax then
	begin
	  Printf.printf "f\n"; (*** this indicates a fundamental problem, not a problem with the ajax input ***)
	  exit 1
	end
      else
	begin
	  Printf.printf "Failure at line %d char %d: %s\n" !lineno !charno x; flush stdout;
	  exit 1
	end

let rec sql_docpresentation docsha c i =
  try
    let l = input_line c in
    sql_docpresentation_0 docsha c l i
  with End_of_file -> ()
and sql_docpresentation_0 docsha c l i =
  if String.length l > 0 && l.[0] = '$' then
    begin
      if String.length l > 1 then
	if l.[1] = 'x' then
	  begin
	    Printf.printf "INSERT INTO `docpresentationpart` (`docsha`,`docpresentationpartno`,`docpresentationpartexercise`) VALUES ('%s',%d,'%s');\n" docsha i (String.sub l 2 (String.length l - 2));
	    if not !presentationonly then (Printf.printf "INSERT INTO `docexercise` (`docsha`,`termid`) VALUES ('%s','%s');\n" docsha (String.sub l 2 (String.length l - 2)));
	  end
	else
	  Printf.printf "INSERT INTO `docpresentationpart` (`docsha`,`docpresentationpartno`,`docpresentationparttreasure`) VALUES ('%s',%d,'%s');\n" docsha i (String.sub l 1 (String.length l - 1));
      sql_docpresentation docsha c (i+10)
    end
  else
    begin
      Printf.printf "INSERT INTO `docpresentationpart` (`docsha`,`docpresentationpartno`,`docpresentationparttext`) VALUES ('%s',%d,'" docsha i;
      try
	sql_docpresentation_1 docsha c l (i+10)
      with End_of_file ->
	Printf.printf "');\n";
    end
and sql_docpresentation_1 docsha c l i =
  if String.length l > 0 && l.[0] = '$' then
    begin
      Printf.printf "');\n";
      sql_docpresentation_0 docsha c l i
    end
  else
    begin
      for j = 0 to String.length l - 1 do
	let z = l.[j] in
	if z = '\'' then
	  (output_char stdout z; output_char stdout z)
	else if z = '\\' then
	  (output_char stdout '\\'; output_char stdout '\\')
	else
	  output_char stdout z
      done;
      output_char stdout '\n';
      let l = input_line c in
      sql_docpresentation_1 docsha c l i
    end

let preset_ptm_lam_id (m:ptm) : string =
  let h = ptm_lam_id m sigtmof sigdelta in
(*  Hashtbl.add sigdelta h m; *)
  h

let preset_ptm_all_id (m:ptm) : string =
  let h = ptm_all_id m sigtmof sigdelta in
  Hashtbl.add sigdelta h m;
  h

let preset_hf_index () =
  Hashtbl.add indextms "73c9efe869770ab42f7cde0b33fe26bbc3e2bd157dad141c0c27d1e7348d60f5" (Ar(Set,Ar(Set,Prop)));
  Hashtbl.add indextms "037e089bccb3925f4dc5460a6a8d3fb7d179f4a65a08d691fef239cef88f89ac" (Ar(Set,Ar(Set,Prop)));
  Hashtbl.add indextms "c271c80f80f5f72a61f48aa63abcf552ccb5c1c1455890804f46f810f52c1725" (Ar(Set,Ar(Ar(Set,Ar(Set,Set)),Prop)));
  Hashtbl.add indextms "b81ebf26fbdeb5d6df6921e54589741b046143546187b8caabb14c3d5592d36a" (Ar(Ar(Set,Ar(Set,Prop)),Prop));
  Hashtbl.add indextms "e4778ebd18e6f80c356efa023ee854f68b4033643e5b0acc995344830efc2fea" (Ar(Set,Ar(Set,Ar(Ar(Set,Set),Ar(Ar(Set,Set),Ar(Set,Set))))));
  Hashtbl.add indextms "d891d45cd1051cf585d9331105c30fa83eb42bb22bc7505aa2482843924b0381" (Ar(Set,Ar(Ar(Set,Set),Ar(Ar(Set,Ar(Set,Prop)),Set))));
  Hashtbl.add indextms "dab6e51db9653e58783a3fde73d4f2dc2637891208c92c998709e8795ba4326f" (Ar(Set,Prop));
  Hashtbl.add indextms "f5bc54642e26d31ca657d19f36547e63838a100f274eeb6bdd98a6e9881d44ab" (Ar(Set,Ar(Ar(Set,Set),Ar(Ar(Set,Ar(Set,Set)),Set))));
  Hashtbl.add indextms "8a8e36b858cd07fc5e5f164d8075dc68a88221ed1e4c9f28dac4a6fdb2172e87" (Ar(Set,Ar(Set,Prop)));
  Hashtbl.add indextms "7b87209a7554e3c82a98275105aedd445c26f03134ed39cc2359a7a68aa5f913" (Ar(Set,Prop));
  Hashtbl.add indextms "d8d907db7e3ec5f56bca5322be5866f518305585f0b078eacd81a019e9c1b007" (Ar(Set,Ar(Set,Ar(Ar(Set,Set),Prop))));
  Hashtbl.add indextms "f36183e461fede37fa3bfca99bb2cf444d783982c4e61ae44ca80c31958792a3" (Ar(Set,Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Ar(Set,Prop))))));
  Hashtbl.add indextms "f95a7cef2d6e940de61c6c672c77c955b8c3f3bc7a1bdc47abe55aeb2471f586" (Ar(Set,Ar(Set,Ar(Set,Prop))));
  Hashtbl.add indextms "2a38dbb37506341b157c11dddf8eb7d8404ce97eb50d5e940b23d5094ae39d70" (Ar(Set,Set));
  Hashtbl.add indextms "5867641425602c707eaecd5be95229f6fd709c9b58d50af108dfe27cb49ac069" (Prop);
  Hashtbl.add indextms "b1b8f6f9d4d6136be8375ed2faddb54df84cdb3018ec6bd2ac07b2c25d3d8af8" (Ar(Set,Ar(Ar(Set,Set),Set)));
  Hashtbl.add indextms "e7493d5f5a73b6cb40310f6fcb87d02b2965921a25ab96d312adf7eb8157e4b3" (Ar(Set,Prop));
  Hashtbl.add indextms "a30e37dacaf00ab138d17f9f913cbfab502311f611c1fba0a4fde0cc946b36ec" (Ar(Set,Prop));
  Hashtbl.add indextms "47ea64150ae653c834316a1ed437b1f7f8f25b8151a683580f54756bd60c3722" (Ar(Ar(Set,Ar(Set,Prop)),Prop));
  Hashtbl.add indextms "77d353d8cc581748150c69e0538465211bee8036c73d724f30d2c874db7fdb2e" (Ar(Set,Ar(Set,Ar(Ar(Set,Set),Prop))));
  Hashtbl.add indextms "a2f75a282012c92e5dcfb6cb7e0391c0c121cea89dcecbe7ef36d5ff8d659654" (Ar(Set,Prop));
  Hashtbl.add indextms "ac96e86902ef72d5c44622f4a1ba3aaf673377d32cc26993c04167cc9f22067f" (Ar(Set,Ar(Ar(Set,Prop),Ar(Ar(Set,Prop),Prop))));
  Hashtbl.add indextms "058f630dd89cad5a22daa56e097e3bdf85ce16ebd3dbf7994e404e2a98800f7f" (Ar(Prop,Prop));
  Hashtbl.add indextms "b8c31f5ff43bf2f656a21709c6a704235abca65899879506dfda2fab358a25c2" (Ar(Set,Ar(Set,Set)));
  Hashtbl.add indextms "e6a3fd57f7f656aedd33ff3147da9d5f8f6797c10f6d29f7fa7b2f9e4ef079e6" (Ar(Set,Set));
  Hashtbl.add indextms "390c91ed9e340848f16dd889ef127c1ce69847c402982071d5840a5b0dfcde97" (Ar(Set,Set));
  Hashtbl.add indextms "ee2e1f36ccc047af9077fcfe6de79d6c9574876b02cae0b4b919e11461760f0d" (Ar(Ar(Set,Ar(Ar(Set,Set),Set)),Ar(Set,Ar(Set,Prop))));
  Hashtbl.add indextms "2f8b7f287504f141b0f821928ac62823a377717763a224067702eee02fc1f359" (Ar(Set,Ar(Set,Prop)));
  Hashtbl.add indextms "d698bec710d5453c25788da55d76957c8215a5706e69835579b33c3ad8eee8c1" (Ar(Set,Ar(Ar(Set,Prop),Set)));
  Hashtbl.add indextms "e3212eac20b484cfefb1aebafa2fad0319a74b904339753a355245c4b597ef51" (Ar(Set,Prop));
  Hashtbl.add indextms "4a0f686cd7e2f152f8da5616b417a9f7c3b6867397c9abde39031fa0217d2692" (Ar(Ar(Set,Ar(Set,Prop)),Prop));
  Hashtbl.add indextms "3bc72e80537e6b6636f39ac493d5dfc92ad4924e70007151a57897df089e61a4" (Ar(Set,Set));
  Hashtbl.add indextms "e2a83319facad3a3d8ff453f4ef821d9da495e56a623695546bb7d7ac333f3fe" (Set);
  Hashtbl.add indextms "54138da5963ad625eb0820ba3d5a7b8124b289d5cdbf3d8e82800f85529293a9" (Ar(Set,Ar(Set,Set)));
  Hashtbl.add indextms "23f36d57906c9e58383c2815f69f0330110ebe6c5f39997a2237e3d50d72756b" (Ar(Set,Prop));
  Hashtbl.add indextms "161886ed663bc514c81ed7fe836cca71861bfe4dfe4e3ede7ef3a48dbc07d247" (Ar(Ar(Set,Ar(Set,Prop)),Prop));
  Hashtbl.add indextms "2452457900954cee0bd913491aa26ba1c9984bafbef8a71bd00fd6b554145708" (Ar(Set,Ar(Set,Set)));
  Hashtbl.add indextms "cfe97741543f37f0262568fe55abbab5772999079ff734a49f37ed123e4363d7" (Ar(Prop,Ar(Prop,Prop)));
  Hashtbl.add indextms "d2a0e4530f6e4a8ef3d5fadfbb12229fa580c2add302f925c85ede027bb4b175" (Ar(Prop,Ar(Prop,Ar(Prop,Prop))));
  Hashtbl.add indextms "548bcb5c04c17b31cc1ec6d6b4b46c678e6e4a357bd25d1695c0bc272ba9a37d" (Ar(Prop,Ar(Set,Ar(Set,Set))));
  Hashtbl.add indextms "4a3cddd8e7c2976b5e5b7ca7da57dca4457ac3d01e2e00e89183a77580d16c1c" (Ar(Set,Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Ar(Set,Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Ar(Set,Set))),Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Ar(Set,Set))),Ar(Ar(Set,Ar(Set,Ar(Set,Set))),Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Prop)))))))))))))));
  Hashtbl.add indextms "374844950523d3fbe55a03d7ae59ea802431e2179d544ee57e31de729819b787" (Ar(Set,Ar(Set,Set)));
  Hashtbl.add indextms "5bf697cb0d1cdefbe881504469f6c48cc388994115b82514dfc4fb5e67ac1a87" (Prop);
  Hashtbl.add indextms "4267a4cfb6e147a3c1aa1c9539bd651e22817ab41fd8ab5d535fbf437db49145" (Ar(Ar(Set,Ar(Set,Prop)),Prop));
  Hashtbl.add indextms "5982716a3cee1f1482d9cbb83c9dcb37b1376aa6a4e1d3b6664c38b9cddc0704" (Ar(Set,Set));
  Hashtbl.add indextms "836ac9a550aae7f3c57da2f661879ac5fe0349f4455ee219d0620ba058b8ab10" (Ar(Set,Ar(Set,Prop)));
  Hashtbl.add indextms "715bbd62fbb17cb65ce6585f3601057f5fa3a5ed09c0b612c375069517ef2518" (Ar(Set,Prop));
  Hashtbl.add indextms "fc5a39eced7dbb49fbfd0a3206540aedab0f9f5e13a0f3900c1c71cbaa7618d0" (Ar(Set,Ar(Ar(Set,Prop),Set)));
  Hashtbl.add indextms "d707030ce5502986326b354994a82ac22f4cf99f23ade56450116039d885a55e" (Ar(Set,Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Ar(Set,Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Ar(Set,Set))),Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Ar(Set,Set))),Ar(Ar(Set,Ar(Set,Ar(Set,Set))),Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Prop)))))))))))))));
  Hashtbl.add indextms "acb8df0c53b21f2b8117b976936b63d805299d2ae2ba2252b959951f7436a198" (Ar(Set,Prop));
  Hashtbl.add indextms "f2b5f44e31fe4f80450c8214164c9d54568b3476dc463bb1b1090e84ad094ff0" (Ar(Set,Ar(Set,Set)));
  Hashtbl.add indextms "0609dddba15230f51d1686b31544ff39d4854c4d7f71062511cc07689729b68d" (Ar(Ar(Set,Ar(Set,Prop)),Prop));
  Hashtbl.add indextms "7da3d166e000be75ef87a4926c1d1e60bb6d72fe624ca2e1896635cac87966e8" (Ar(Set,Prop));
  Hashtbl.add indextms "70556f3106ecb4913c6bbf5b05f6723b56f418eb0d4f08c7e057e55c68c0bd41" (Ar(Set,Ar(Set,Prop)));
  Hashtbl.add indextms "6953ff7fbdface4e2c4f2fc7119981e1ec6b514dc3f2f40ff3a75135974c42b1" (Ar(Set,Set));
  Hashtbl.add indextms "dc688de6dbfa5c75bd45f1eb198583d07be144f0cdabb44def09da1c0976495b" (Ar(Set,Set));
  Hashtbl.add indextms "844774016d959cff921a3292054b30b52f175032308aa11e418cb73f5fef3d54" (Ar(Set,Set));
  Hashtbl.add indextms "5dafc57212bbd767c604798a3851a9a73727a8c7f472ea9676ab306063b9a5d6" (Ar(Set,Ar(Ar(Set,Prop),Ar(Ar(Set,Prop),Prop))));
  Hashtbl.add indextms "e94e680d80fd51ff5fcacd526ea5e7c772160c7f0463695723da0c525aac3215" (Ar(Set,Ar(Set,Set)));
  Hashtbl.add indextms "44142c53f9949e926aafaef97c1eea494a8bccf3d72c7d8eec1c6b798276c075" (Ar(Ar(Set,Ar(Set,Prop)),Prop));
  Hashtbl.add indextms "67bdfa716c9dcdd80aa90d444691004d697518dbc44d860bc7fe912388f4f53a" (Ar(Set,Ar(Ar(Set,Ar(Set,Set)),Ar(Set,Set))));
  Hashtbl.add indextms "35f61b92f0d8ab66d988b2e71c90018e65fc9425895b3bae5d40ddd5e59bebc1" (Ar(Ar(Set,Ar(Set,Prop)),Prop));
  Hashtbl.add indextms "7e180af329e6cf47cfc0f1e1b6f255a489e841155a7ef45036ab35219ecc394a" (Ar(Set,Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Ar(Set,Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Ar(Set,Set))),Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Ar(Set,Set))),Ar(Ar(Set,Ar(Set,Ar(Set,Set))),Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Ar(Ar(Set,Ar(Set,Set)),Prop)))))))))))))));
  Hashtbl.add indextms "26f19814fc7047e59d540f748532067762b0b3fe77962c9a3cf69d22b46315da" (Ar(Ar(Set,Ar(Ar(Set,Prop),Prop)),Ar(Set,Ar(Ar(Set,Prop),Prop))));
  Hashtbl.add indextms "d772b0f5d472e1ef525c5f8bd11cf6a4faed2e76d4eacfa455f4d65cc24ec792" (Ar(Set,Ar(Ar(Set,Set),Set)));
  Hashtbl.add indextms "591ebe2d703dc011fd95f000dd1ad77b0dca9230146d2f3ea2cb96d6d1fba074" (Ar(Ar(Set,Ar(Set,Prop)),Prop));
  Hashtbl.add indextms "0dc5b3de8b5f01c0ca9c1c95a17181420737811449cf8e68feb2367adeff06db" (Ar(Set,Set));
  Hashtbl.add indextms "1c3c2c8b70491b0b7719dee35a9acf68014c4ee03d4c9c85959529eddff31afe" (Ar(Set,Prop));
  Hashtbl.add indextms "8f39e0d849db8334a6b514454a2aef6235afa9fc2b6ae44712b4bfcd7ac389cc" (Ar(Ar(Set,Ar(Set,Prop)),Prop));
  Hashtbl.add indextms "6f17daea88196a4c038cd716092bd8ddbb3eb8bddddfdc19e65574f30c97ab87" (Ar(Set,Ar(Set,Ar(Set,Prop))));
  Hashtbl.add indextms "de2a1ef5a65d0333bfbd45429b1eeb5a777fac1eedaff25d7e927fbb05864e04" (Ar(Set,Ar(Set,Prop)));
  Hashtbl.add indextms "fe3392384b653ab493e6f67711d18e44f10327d41d3123c90a95d5c79ce1010c" (Ar(Set,Ar(Set,Set)));
  Hashtbl.add indextms "4d24e53133cd0703a4721d279e15ca1aa7ca5645d1d8ae6292d11ce75b844de0" (Ar(Set,Ar(Set,Set)));
  Hashtbl.add indextms "fd305ee0551a14932538dca1789e7c3960fb5751bc50fd6b5feae33060791cbc" (Ar(Set,Ar(Set,Set)));
  Hashtbl.add indextms "8dadc51fb1798533b4f8b517aa2e027c4aba5de5c0dafe69a8fe198b6f0692bc" (Ar(Set,Set));
  Hashtbl.add indextms "9c60bab687728bc4482e12da2b08b8dbc10f5d71f5cab91acec3c00a79b335a3" (Ar(Prop,Ar(Prop,Prop)));
  Hashtbl.add indextms "9647b545c9680812ae9fd2e1b0809fd3aea9c3573aa45a57bb969e7445cd0bbf" (Ar(Set,Ar(Ar(Set,Prop),Ar(Ar(Set,Set),Set))));
  Hashtbl.add indextms "87fba1d2da67f06ec37e7ab47c3ef935ef8137209b42e40205afb5afd835b738" (Ar(Prop,Ar(Prop,Prop)));
  Hashtbl.add indextms "e8781715d856afa33ba260fc3b89715e4e75c58ffd89f290eb4199476d0f48c0" (Ar(Set,Prop));
  Hashtbl.add indextms "b617d031e612cb38cbe55ad4b5ecd6773e49d1fdddf9c166e959483d11ce24a0" (Ar(Set,Ar(Set,Set)));
  Hashtbl.add indextms "174b78e53fc239e8c2aab4ab5a996a27e3e5741e88070dad186e05fb13f275e5" (Ar(Ar(Set,Prop),Set));
  Hashtbl.add indextms "cc1c6dfc5df6c4fa4241c37b21da1d6667759fc582b9cb4024730dec0b7b65b8" (Ar(Set,Prop));
  Hashtbl.add indextms "9b5b8829613dd319c4e49f7f0685e927870a2d639b99cc6868b31706e1f6bb63" (Ar(Set,Prop));
  Hashtbl.add indextms "3578b0d6a7b318714bc5ea889c6a38cf27f08eaccfab7edbde3cb7a350cf2d9b" (Ar(Prop,Ar(Prop,Prop)));
  Hashtbl.add indextms "b4074dc0a3edfa313765b648763f63a0f2e0852a1a6fed8d34a3a95047e77b25" (Ar(Set,Prop));
  Hashtbl.add indextms "3f9587947480129a19f5340fc9f47f92a36f268cd54789bc27e5b79f4f7d133e" (Ar(Set,Ar(Set,Prop)));
  Hashtbl.add indextms "cd593a5b8620236c9a1ddf89ecf85e4bb85fcdcf351d9d390186775a6189a278" (Ar(Set,Ar(Set,Set)));
  Hashtbl.add indextms "eb13782dca61e45f1b194346d4ef12ef9f40ce69b8cd09ed87242557c338e8ce" (Ar(Set,Ar(Ar(Set,Set),Set)));
  Hashtbl.add indextms "1a7d5254d6fa7bc2304a87955df78ab68aa34b8288c14e27183859c17a5af3a6" (Ar(Set,Ar(Set,Prop)));
  Hashtbl.add indextms "939c8985880c5d4a723190f9c3c3605d3fd4e502ec46d099a4a4cbd4b6559135" (Ar(Set,Ar(Ar(Set,Prop),Ar(Set,Ar(Ar(Set,Prop),Prop)))));
  Hashtbl.add indextms "8a011777f9521bcd1d03f0c8b7103df4392c97aa5ab5e03bcc7e774ca01f8f53" (Ar(Ar(Set,Ar(Set,Prop)),Prop));
  Hashtbl.add indextms "1ecbb47bb099cec6d8abb8754adbc5da7102f52677e6311a3b1c5e310424535e" (Ar(Set,Ar(Set,Prop)));
  Hashtbl.add indextms "c745d42012a6c8c4ccc0ecacdc4c543f1b028344a4589ccb30f19e6563441a39" (Ar(Set,Prop));
  Hashtbl.add indextms "b74baf978a7834845e2cb5395f01c98d61dca8eeb682ec1717797bda646d34da" (Ar(Set,Ar(Set,Set)));
  Hashtbl.add indextms "fe4c3851a2360b02b580e3ee88734f04b75caf81640085091699e5107ffe00be" (Ar(Set,Prop));
  Hashtbl.add indextms "3e5bc5e85f7552688ed0ced52c5cb8a931e179c99646161ada3249216c657566" (Ar(Ar(Set,Ar(Set,Prop)),Prop));
  Hashtbl.add indextms "52dcdfd944c55d489f20fe755a6f20841ae9daa9e7816972f530a28055a04466" (Ar(Set,Set));
  Hashtbl.add indextms "a570d71cf4f89199d200ff1908f6dc1f95baec3e9f9e5b7f7756d8488c16ae7e" (Ar(Ar(Set,Ar(Set,Prop)),Prop));
  Hashtbl.add indextms "dbd86f850e734b76bb1d9f35a0d37ee55973f6af1afabcebd9f7cb4edc2f5970" (Ar(Set,Set));
  Hashtbl.add indextms "8ec8f2cdfc2de7751ed810752e14b4b39c9495765ecfd5410f1437a905135b61" (Ar(Set,Ar(Ar(Set,Set),Set)));
  Hashtbl.add indextms "556f5d266d77e6dfcc067c1f9cc93c5522c6e2b1740eb7ce485c28e2ac69efd7" (Ar(Ar(Set,Prop),Prop));
  Hashtbl.add indextms "d1aaa38ae00af3aed91194e61473f572356d230c7e7a05508b0fe173b378fc62" (Ar(Set,Ar(Ar(Set,Prop),Ar(Set,Ar(Ar(Set,Prop),Prop)))));
  Hashtbl.add indextms "f275e97bd8920540d5c9b32de07f69f373d6f93ba6892c9e346254a85620fa17" (Ar(Set,Ar(Set,Prop)));
  Hashtbl.add indextms "95dfcd440145c3d7a84fe79b558820a60d3d34cb961a0144e83436d21840852b" (Ar(Set,Prop));
  Hashtbl.add indextms "f97da687c51f5a332ff57562bd465232bc70c9165b0afe0a54e6440fc4962a9f" (Ar(Ar(Set,Ar(Ar(Set,Set),Set)),Ar(Set,Set)));
  Hashtbl.add indextms "4d11922bee7b3b3664befde98c70798021a1da8be65cc37a9e06f634724d5adb" (Ar(Ar(Set,Ar(Ar(Set,Prop),Prop)),Ar(Set,Ar(Ar(Set,Prop),Prop))));
  Hashtbl.add indexknowns "868294bdc937d049c87087910f30abde9c34a8b67f658f43eb55497ee4fdf496" ();
  Hashtbl.add indexknowns "4e2bc7210cca188ef72b6d1b20c4c39a92ac7ec6125ddca771f7a09009ead068" ();
  Hashtbl.add indexknowns "15fec847b5eedab6f8bc99ebb50781626caef5d65b4ef634ac1c140003145eba" ();
  Hashtbl.add indexknowns "2599898996462cb0162aff0720f0d4eb974df2d07bc5fb8eeaf0fc413dea29ef" ();
  Hashtbl.add indexknowns "dbd431814e07ea423fc6370b502780c8a0872537fd7a4c5bd4549246503ae3ed" ();
  Hashtbl.add indexknowns "f025439203fb01208b8741e899c3517dde02d78a3ac94c10d78cc84af88508bd" ();
  Hashtbl.add indexknowns "365a7281af1cbf5fc69c86c6e7ba6645ec88e69ee08fa41cb3507cbb003d4eeb" ();
  Hashtbl.add indexknowns "04066c2662d563918cacf3cdb8201eadccda7b3cd977dfb5c48f0ae0c4404814" ();
  Hashtbl.add indexknowns "b200928a20d21735cc0af6fa7c870b170a4db3d5e88217cc242329b25d84cf86" ();
  Hashtbl.add indexknowns "464987a2ad498860dd6a7cc33bada9e963b7bb7853b08edfd85f53b6f6fe7381" ();
  Hashtbl.add indexknowns "f7a09b7a58d286a0dfb678d8da4ab8320983e945baf203133136ac4d50401b41" ();
  Hashtbl.add indexknowns "12820df3302265c7fbaf0a4fbe09ac48b4ebad48e7953a671aee16d60a5610f3" ();
  Hashtbl.add indexknowns "55a742b445a6c791407a41bcacf3cb7b1ec710a463c251f6134eaa2de74bcf18" ();
  Hashtbl.add indexknowns "cfbc835ac32334e3211b53cf4a1148efd492b9823fece69209e1f2887f4f57ff" ();
  Hashtbl.add indexknowns "d2009e03cbbe1ff649d90ca190353c0195ad2a37a3e0b4aae69e88f5760763b4" ();
  Hashtbl.add indexknowns "224f062ec0e5871d4e35b2277008bcbb5cf06f9cee2a8c0d0f3797dfafc390e3" ();
  Hashtbl.add indexknowns "c6f87070e0538a2770d3dd759caea31f3628afbec8831176292f6ea08b25af4f" ();
  Hashtbl.add indexknowns "8c71cf7ec1276527982a30a7e0cf59ab0066fe232b08a93bde0e96f04bd3c61e" ();
  Hashtbl.add indexknowns "7c67e84f3df4606d476677f0a80b26df40de299f16e87eafc3d28a86de41d745" ();
  Hashtbl.add indexknowns "6601a8127d457a3fdaaae9564da5f2b6143ce5ae9efd358aa178f842fdaae6a8" ();
  Hashtbl.add indexknowns "41b1004b8f2e1e9141952323c57666ef1579b53152eaac85192b8cb654c248b2" ();
  Hashtbl.add indexknowns "1174f35d06d32f6ce8764ddf4bb7770c276872fa330e758d502bfb31a70367f5" ();
  Hashtbl.add indexknowns "5206cb1e5c68cffe15ac7d921c8f21359f42b338372c195a95adce69872cfe55" ();
  Hashtbl.add indexknowns "2d84495a5a228c61e5bcfa757b06de12a5cb7eaf22e1979742846f5d18a7a98d" ();
  Hashtbl.add indexknowns "0fab0105f86e62d59255f98965bf1f26e395b983e8db295c63f7ead552040aef" ();
  Hashtbl.add indexknowns "a703842a9d0a604aa91bd6de73ae7c7599f2b89d0195246be3eb7cfad204f480" ();
  Hashtbl.add indexknowns "6102787581a22e0f58454da3c71dd11253824a00fbd97b9e4f475a608e2018dc" ();
  Hashtbl.add indexknowns "87939657b3fe4d552dea1c942c9f6003a7fcae766448ff94352c3386da905557" ();
  Hashtbl.add indexknowns "572abf072260ddf309f0a2f3b3f13276fefde702e2f5c70aaa72cc2efbb816ef" ();
  Hashtbl.add indexknowns "dc1664fc81c300e53160f04faaeaff65a2bb7223fee8d688ae4138c009f113dd" ();
  Hashtbl.add indexknowns "5a3fd8f2cf1480eee62b842b1c0b3b6db5ba80c1985063ca6dcbb06abf4d34f0" ();
  Hashtbl.add indexknowns "b492ab96942311595fd53c93243cb7ab5314986bb8460d580d9382dfab90f7d1" ();
  Hashtbl.add indexknowns "82782889d9aa2bb5bf98d2b277bf4dd619fc1ddaf5b3122f83fe3b8409402109" ();
  Hashtbl.add indexknowns "2e048bda28405de21b9e617b3f745e5d3035361962c675a82a3b5b0bab8d9300" ();
  Hashtbl.add indexknowns "586cb383efcc9274c1a5cd6c007638d11fafffa82c9ca73a3b2df8bbea8a0ac7" ();
  Hashtbl.add indexknowns "867da1c87d86c7decce481bc3ce02387488e600047c980a9987e5d3a84819255" ();
  Hashtbl.add indexknowns "99dcd262fd027f306fbe8ad22cbd1732f1179e836383110a981b77b6bc2dbeac" ();
  Hashtbl.add indexknowns "2258ad14aeb3646753a99caf86749fafbed7e4c6b96b123c179b5eb35590c00c" ();
  Hashtbl.add indexknowns "d5fdb4f6cfb82cab64716bee0629544da9b7530752eb0873529def98362fd6b4" ();
  Hashtbl.add indexknowns "13258e1f5d4a2643a1549f9714a456fca2aade1be40f8499faa7cd1c4869d332" ();
  Hashtbl.add indexknowns "54b3761cc4b2183c3459b377fb2511e605f1a8fd2f1ddebbfc909450dad5a4ef" ();
  Hashtbl.add indexknowns "f527171847882df226b7124e83d0030a8d25a522554ec88ebdfdf6f3f4d0f0e8" ();
  Hashtbl.add indexknowns "7f6246d08629eeb16eab93529ffe4f929f43344833ab88c7786393693520e82b" ();
  Hashtbl.add indexknowns "4cc5a2ff81402353f76b313a8c49f54c017dda5deb9df22bac241e5b1f3edc30" ();
  Hashtbl.add indexknowns "e003a9ce9f2518b46b0d1c2c7cd8d4dd040b23d00931d08136c96df181ef6979" ();
  Hashtbl.add indexknowns "6660815c9e18e7c5d06d6b80ed5861acf153606a08bd14cb03f5ce1f9bb7c549" ();
  Hashtbl.add indexknowns "df80873a64df56bcf11135a2789ea9b13989067b38647620bb14a7002ebe7bbf" ();
  Hashtbl.add indexknowns "3420d9d032cef82dd17e9dacfe93f8763ff4470d6c5faab0c94d9b692148c412" ();
  Hashtbl.add indexknowns "626dea760cd96620e6463e4a06173c0cb8f75c53c08b67df10983c723681d836" ();
  Hashtbl.add indexknowns "96d4dd0c1ba9b65eecf037e4e8d9b8357f903b6a6e7b0d5840b34402a761e2aa" ();
  Hashtbl.add indexknowns "554a477931702afb1eb68ab4324d4de9da80e73236430fade24c2c7080cfb1fa" ();
  Hashtbl.add indexknowns "8725773844f3b041309396003ac8e76b21fcb56c96e54747e95ba01a831f33e8" ();
  Hashtbl.add indexknowns "162375fbd86c5d519196cf18da14d5ed54c3f7cf82940247e45769aa76a57e40" ();
  Hashtbl.add indexknowns "ac246827b8e30ee0bdb769ad911ccc2c1e1a54654ecaa200e86fbb3da01a6a61" ();
  Hashtbl.add indexknowns "52482519eeceb4095a9410aafda54d9170e6251f2659499c52424539b6ba90a9" ();
  Hashtbl.add indexknowns "4a3f268d85564bb8791305d5cfc00c7ca583ac2643af08c9c64aaed9d9502d2b" ();
  Hashtbl.add indexknowns "d2f3d81ce8b6ea6e95412f66720543226e9ea3741ce88e14e8e3c4c8b7812a0a" ();
  Hashtbl.add indexknowns "f1d38952a8e7f77cb5b871bea26c16ee679f1bf1bd1bc40c7985e7187c3ea0bb" ();
  Hashtbl.add indexknowns "d587051f0ca663beaee539b5c7c1a03ceaa1b46c97d7383a1ca8806db09f1d8e" ();
  Hashtbl.add indexknowns "366c5817fcc8d04be3dca45a9e89b91d6aade49505bc119294c3de0c7d18e564" ();
  Hashtbl.add indexknowns "de0eaf5d33c0b573d63cc09df92572f243045babd3e3cb9aa49e72d618f6c7c9" ();
  Hashtbl.add indexknowns "82df04260594aa489dc5962975a15987938bbb37690535eaaeaa91b7c08a3868" ();
  Hashtbl.add indexknowns "6620b2ab1330a972ee8d13edfb67de4eacfea4623021a69ab91aba1bc5486482" ();
  Hashtbl.add indexknowns "00821ffdcd36dda11c2adb0d9ba6d074888e7e61d5c48a8edf3d5c49d89326b3" ();
  Hashtbl.add indexknowns "5ee4a4103f04cabe781fcdc73566d7dd74b33cb621a83145e1fcff8855469827" ();
  Hashtbl.add indexknowns "64272f2d4ac604200cce5a7199adacfe999b58f69968fab44d934b23391b77eb" ();
  Hashtbl.add indexknowns "b35f97e058888684b8c42a84bc499032f00499ebcbd1102d4e5126243eaf92f7" ();
  Hashtbl.add indexknowns "f90aa94ea2ccb281dc50b80c53cd84d43c275f66c3cb45ef83130cd5ea5f84db" ();
  Hashtbl.add indexknowns "ee04f5da0f3d645ba581e47989d64793ffb8cd3fdc8a053aa7f70c9f23c8df23" ();
  Hashtbl.add indexknowns "6106bdc32891cfc869e9ed570bcfcc8a710c526148797a8e3016d1d71cc0ae4d" ();
  Hashtbl.add indexknowns "e788b52757ccbf1862b39f5567d43501ac796f21bffe2ea598d811bbdb2133d3" ();
  Hashtbl.add indexknowns "692ba143db461ed1eecaf80a8d1b3a764e60a1bba5e63aded7267c4c9296cd9d" ();
  Hashtbl.add indexknowns "39d403033e64174a835078c562805e92c60140a2c89cfb32806fbf1eb4ed7022" ();
  Hashtbl.add indexknowns "ef3aa3e1b4d689ddd4c0fe2668544fdb470034e66da44594d21a07ac7aed20da" ();
  Hashtbl.add indexknowns "bd76e531fc7a8ed39d008c4f87c276d214b522875c8315b99b6c9b38485023d2" ();
  Hashtbl.add indexknowns "dc87a9ffd7a39a5d72301ac394d9604bbdfb6b2cfe5e507a7248d05a87ad6f60" ();
  Hashtbl.add indexknowns "9e462710824cabfec9de39cdff20d19402310de37dfef2dc568e559856c868aa" ();
  Hashtbl.add indexknowns "f37386d70cae9fddb0dc5ca7e4ce6be135d7ac249832b637f6adfdd8b801b5e3" ();
  Hashtbl.add indexknowns "064251eb4d64bfcfe2dbb7036e2cffb5e9be4debe64169fc0b64e5c816af2d2d" ();
  Hashtbl.add indexknowns "866dbcf42d4554c0d35bc97d0273c3851c37e26e2a87fa1e029b654d8bf8d1ca" ();
  Hashtbl.add indexknowns "27759e669158c829fb7d3d5b7f08a9a556dace9913d27ba8a473751e0f58a125" ();
  Hashtbl.add indexknowns "8f7fc690f78c58bbd65abb579da0a2ea977736fa04f68db517c1c38b08b99db0" ();
  Hashtbl.add indexknowns "7952f08fb7fe7fb44139cd3db3228728b70d916b37bd8b042cea4808b7f23407" ();
  Hashtbl.add indexknowns "9d44a6b7062eb0daf6dee4f01808066e44babe00dab2f300d5c2888e2373aac3" ();
  Hashtbl.add indexknowns "ef0fdb780cd7881d61cdf680d4163acad949a5d9ceb5f69bbfe03e6b7e1e13aa" ();
  Hashtbl.add indexknowns "3443f52b74e17e8105890fbeac1841b246d71469124783f7c9fdadc8b49308be" ();
  Hashtbl.add indexknowns "53d1c88cf93b204c23fdc73ee1f2a31faab3555092dcf2d13bd2930ade3aa9fd" ();
  Hashtbl.add indexknowns "1ee3302698bcb2c7867b8148330fec36841a07ec0b898f45cb8df19bd47ad8ca" ();
  Hashtbl.add indexknowns "8d5eb1ebcd652bb09f43c4e1195948a20678aa89140a42013186ec6bc4e0e0cc" ();
  Hashtbl.add indexknowns "4ef4fc126a12b2291e1d600f286444a9a3606d764e477007b40b95371bf7662c" ();
  Hashtbl.add indexknowns "eaaf4094c0ea89ec07b329573cb8bedd33850f5653f1510b2f843cdbdaceb9f7" ();
  Hashtbl.add indexknowns "abaac70d1f1b814c8db56598d6add19e82cd31fbb80b6d00d5ba992524854b0c" ();
  Hashtbl.add indexknowns "7150dcb5337f259199e19ef556b5bb9b251a47e4a5a86cd0da160c3a54f2e681" ();
  Hashtbl.add indexknowns "7f1c52313b1eeccaf710e7b8ebe2bcd81bca044598e4bfb524cc12ea41ce29e7" ();
  Hashtbl.add indexknowns "4d0b1b3ca489073915e8701e021a09da22887f41f83e407df8eafe040825db18" ();
  Hashtbl.add indexknowns "b20fe08e4deecde7da240694359a1d03879d693ecedd05b8558b7c8f97a27068" ();
  Hashtbl.add indexknowns "1b21a591a91c0eb2f7d4082a3c5bf9588c29e41b071c31e04a33ebb58a8dda8d" ();
  Hashtbl.add indexknowns "f5c78d7ac16bbd897ebb3d454ab1cd7f9293aacb85b51d6e2d77ac161c6aeaa9" ();
  Hashtbl.add indexknowns "604c7a9ddc69ab30c7a270d93252bf0a28677a3c64eebd21010fa447cb8add57" ();
  Hashtbl.add indexknowns "294aa57cc6984370c318cea497711321b40c0f0b05ee524a6d5aa73e794e39ea" ();
  Hashtbl.add indexknowns "5cba52f2347cd25c57321e4b4b84c86be9c4563eab6471d084a316fba6d25030" ();
  Hashtbl.add indexknowns "d8b1e4f182bfcec470fe7eea737c79a527fc286ab08d2c395b0361f45f2ba0ee" ();
  Hashtbl.add indexknowns "741acbc5a4bf92baedfa5c23ed298e1b520b91aed70ca83f4269f898a4106541" ();
  Hashtbl.add indexknowns "7b2e02f7b16b4025d23cc225cb002a66efbb3e6e8cdd4e9414f2f95aa5f6f1f4" ();
  Hashtbl.add indexknowns "141dce782ecd197c13dc05628f4fe040032438393762d62874daffe79ef2268a" ();
  Hashtbl.add indexknowns "e4b03c310442ae760be9945e176494db51515dfb952ee60fbd42e05527752af0" ();
  Hashtbl.add indexknowns "d19695f3f232d868e0d173308b1485f5a672db8d6d2e2fdb22798c72e3a24b1e" ();
  Hashtbl.add indexknowns "b10119e58d408c73522a7820cd9a29b903ddc3a37b5fe8fca1e45bd1b4101b5f" ();
  Hashtbl.add indexknowns "7d6b2d8ec78bd51bbac37f07461ac3e1e253da81294eca51f8eadb3857ea1baa" ();
  Hashtbl.add indexknowns "6ba14c122e9a0dc11fb0cc698721916647647d29869c54a909f7ac9aaa20b4de" ();
  Hashtbl.add indexknowns "6b83e25db3df911537a2cde43995008632e245b0679e12c1949fb168ef626372" ();
  Hashtbl.add indexknowns "bfd151f806ea28dc7a1dc0837502b7ceb6f8c9d6f8ca53206a01a29e23c88acf" ();
  Hashtbl.add indexknowns "c3226671e2aa9c97d68996272ba2f85e00d1e4985718658f39eb97f95cf854a8" ();
  Hashtbl.add indexknowns "904df15a641757811fc64e68c8790dc99255f6f23a5a1fca8a2dd0062214dd5f" ();
  Hashtbl.add indexknowns "3ab62e54b8183ce5a71e0795deda20c0690cfd0e9ebb775b90c0af1dd485b602" ();
  Hashtbl.add indexknowns "b86970ee1757eea1f06bfdbeab04661f2eb58efecec3ec51a690f2034fb178a2" ();
  Hashtbl.add indexknowns "7b7a158582a755e85390b8b837c8791550dc4bf714a197993bdb4071ed9f4bdb" ();
  Hashtbl.add indexknowns "7d00bc10f6ad193d17d0b88423d050d76590ec9a00f9b240e0f2ccf907eee3ed" ();
  Hashtbl.add indexknowns "ff656464c2a35843ba4bbf94ca8fcda010281b38ff3033ab11c801132ecf4488" ();
  Hashtbl.add indexknowns "36cbd0e8c9859edf89f9d4e1c909caa30ef1b164cf86c3a590cd13c7750534e9" ();
  Hashtbl.add indexknowns "73fa935250c44cafb3971bffe3b0cdef8869cd5693da6149f81741261543e04d" ();
  Hashtbl.add indexknowns "922c01fecc8a6d771b07a08c465fb687f099d674c3630deb232d5553ec648ad1" ();
  Hashtbl.add indexknowns "a568f6925326a250e7600b97f266e41dc669ec030235bc057dd16ef29828e7e4" ();
  Hashtbl.add indexknowns "b473ad9c9918cc7be1d7c29b9ee03bc23f47466c37aeaf71e6be8cb7e1cf303f" ();
  Hashtbl.add indexknowns "dd0ad40c7379761f12272cbe82205119e86020fa48b5cb68a84e812ea0dc4cf2" ();
  Hashtbl.add indexknowns "afec65b390095d370779e024c711ab868bf135e8a62cdf3b8893e70afc7503d5" ();
  Hashtbl.add indexknowns "7e67c13b91338ac608b29691e8375854b932c9b06d10cacb0c9e83d08478eb1a" ();
  Hashtbl.add indexknowns "e92a7f5e5a098d789a262e5d6ee3ab0e4695848c592079f62bbd5bf43c2872c4" ()

let preset_mizar_index () =
  Hashtbl.clear indextms;
  Hashtbl.clear indexknowns;
  (** In is already Prim 1 and Union is already Prim 3 **)
  Hashtbl.add indextms "174b78e53fc239e8c2aab4ab5a996a27e3e5741e88070dad186e05fb13f275e5" (Ar(Ar(Set,Prop),Set));
  Hashtbl.add indextms "73c9efe869770ab42f7cde0b33fe26bbc3e2bd157dad141c0c27d1e7348d60f5" (Ar(Set,Ar(Set,Prop)));
  Hashtbl.add indextms "f55f90f052decfc17a366f12be0ad237becf63db26be5d163bf4594af99f943a" (Ar(Set,Ar(Set,Set)));
  Hashtbl.add indextms "844774016d959cff921a3292054b30b52f175032308aa11e418cb73f5fef3d54" (Ar(Set,Set))

let preset_hoas_index () =
  Hashtbl.clear indextms;
  Hashtbl.clear indexknowns;
  Hashtbl.add indextms "d58762d200971dcc7f1850726d9f2328403127deeba124fc3ba2d2d9f7c3cb8c" (Ar(Set,Ar(Set,Set)));
  Hashtbl.add indextms "73c9efe869770ab42f7cde0b33fe26bbc3e2bd157dad141c0c27d1e7348d60f5" (Ar(Ar(Set,Set),Set))

(*** "main" ***)
let _ =
  (*** There are some global names I need before getting started to make the proof tactics work, so they are precomputed here. ***)
  fal := preset_ptm_lam_id (0,All(Prop,DB(0)));
  fale := preset_ptm_all_id (0,Imp(TmH(!fal),All(Prop,DB(0))));
  eqPoly := preset_ptm_lam_id (1,Lam(TpVar(0),Lam(TpVar(0),All(Ar(TpVar(0),Ar(TpVar(0),Prop)),Imp(Ap(Ap(DB(0),DB(2)),DB(1)),Ap(Ap(DB(0),DB(1)),DB(2)))))));
  conj := preset_ptm_lam_id (0,Lam(Prop,Lam(Prop,All(Prop,Imp(Imp(DB(2),Imp(DB(1),DB(0))),DB(0))))));
  disj := preset_ptm_lam_id (0,Lam(Prop,Lam(Prop,All(Prop,Imp(Imp(DB(2),DB(0)),Imp(Imp(DB(1),DB(0)),DB(0)))))));
  expoly := preset_ptm_lam_id (1,Lam(Ar(TpVar(0),Prop),All(Prop,Imp(All(TpVar(0),Imp(Ap(DB(2),DB(0)),DB(1))),DB(0)))));
  expolyI := preset_ptm_all_id (1,All(Ar(TpVar(0),Prop),All(TpVar(0),Imp(Ap(DB(1),DB(0)),Ap(TpAp(TmH(!expoly),TpVar(0)),DB(1))))));
  preset_hf_index();
  let i = Array.length Sys.argv in
  if i = 1 then
    mgcheck stdin (*** if no arguments are given, read and check from stdin ***)
  else (*** otherwise, assume the last argument is the main file from which to read and check ***)
    begin
      (*** Process command line arguments including reading signature input files ***)
      let j = ref 0 in
      while (!j < i - 2) do
	incr j;
	if Sys.argv.(!j) = "-I" then
	  begin
	    match !latex with
	    | Some(_) -> raise (Failure("-I must come before -latex"))
	    | None -> includingsigfile := true
          end
        else if Sys.argv.(!j) = "-latex" then
          begin
	    if !j < i-2 then
	      begin
		incr j;
		latex := Some(open_out (Sys.argv.(!j)))
	      end
	    else
	      raise (Failure("Expected -latex <filename>"))
          end
	else if Sys.argv.(!j) = "-s" then
	  begin
	    includingsigfile := false;
	    if !j < i-2 then
	      begin
		incr j;
		match !sigoutfile with
		| Some _ -> raise (Failure("Cannot use -s twice"))
		| None -> sigoutfile := Some (open_out (Sys.argv.(!j)))
	      end
	    else
	      raise (Failure("Expected -s <outsigfilename>"))
	  end
        else if Sys.argv.(!j) = "-mizar" then
          begin
            pfgtheory := Mizar;
            preset_mizar_index();
          end
        else if Sys.argv.(!j) = "-hoas" then
          begin
            pfgtheory := HOAS;
            preset_hoas_index();
          end
	else if Sys.argv.(!j) = "-pfg" then
	  begin
	    pfgout := true;
	  end
	else if Sys.argv.(!j) = "-indout" then
	  begin
	    includingsigfile := false;
	    if !j < i-2 then
	      begin
		incr j;
		match !indoutfile with
		| Some _ -> raise (Failure("Cannot use -indout twice"))
		| None -> indoutfile := Some (Sys.argv.(!j))
	      end
	    else
	      raise (Failure("Expected -indout <indoutfilename>"))
	  end
	else if Sys.argv.(!j) = "-ind" then
	  begin
	    includingsigfile := false;
	    if !j < i-2 then
	      begin
		incr j;
		let c = open_in (Sys.argv.(!j)) in
		read_indexfile c;
		close_in c;
	      end
	    else
	      raise (Failure("Expected -ind <indexfilename>"))
	  end
	else if Sys.argv.(!j) = "-ownedout" then
	  begin
	    includingsigfile := false;
	    if !j < i-2 then
	      begin
		incr j;
		match !ownedoutfile with
		| Some _ -> raise (Failure("Cannot use -ownedout twice"))
		| None -> ownedoutfile := Some (Sys.argv.(!j))
	      end
	    else
	      raise (Failure("Expected -ownedout <ownedoutfilename>"))
	  end
	else if Sys.argv.(!j) = "-owned" then
	  begin
	    includingsigfile := false;
	    if !j < i-2 then
	      begin
		incr j;
		let c = open_in (Sys.argv.(!j)) in
		read_ownedfile c;
		close_in c;
	      end
	    else
	      raise (Failure("Expected -owned <ownedfilename>"))
	  end
	else if Sys.argv.(!j) = "-solves" then
	  begin
	    includingsigfile := false;
	    if !j < i-2 then
	      begin
		incr j;
		solvesproblemfile := Some (Sys.argv.(!j))
	      end
	    else
	      raise (Failure("Expected -solves <indexfilename>"))
	  end
	else if Sys.argv.(!j) = "-reportpfcomplexity" then
	  begin
	    includingsigfile := false;
	    reportpfcomplexity := true
	  end
	else if Sys.argv.(!j) = "-reporteachitem" then
	  begin
	    includingsigfile := false;
	    reporteachitem := true
	  end
	else if Sys.argv.(!j) = "-webout" then
	  begin
	    includingsigfile := false;
	    webout := true;
	  end
	else if Sys.argv.(!j) = "-ajax" then
	  begin
	    includingsigfile := false;
	    if !j < i-2 then
	      begin
		ajax := true;
		incr j;
		ajaxpffile := Sys.argv.(!j);
		verbosity := 0;
	      end
	    else
	      raise (Failure("Expected -ajax <pffile>"))
	  end
	else if Sys.argv.(!j) = "-sqltermout" then (*** this should only be used when -sqlout is already being used ***)
	  begin
	    includingsigfile := false;
	    sqltermout := true;
	  end
	else if Sys.argv.(!j) = "-presentationonly" then
	  begin
	    includingsigfile := false;
	    presentationonly := true;
	  end
	else if Sys.argv.(!j) = "-thmsasexercises" then
	  begin
	    verbosity := 0;
	    includingsigfile := false;
	    thmsasexercises := true;
	  end
	else if Sys.argv.(!j) = "-v" then
	  begin
	    if !j < i-2 then
	      begin
		incr j;
		verbosity := int_of_string (Sys.argv.(!j))
	      end
	    else
	      raise (Failure("Expected -v <verbositynumber>"))
	  end
	else if !includingsigfile then
	  let c = open_in (Sys.argv.(!j)) in
	  mgcheck c;
	  title := None;
	  authors := [];
	  includedsigfiles := Sys.argv.(!j)::!includedsigfiles
	else
	  raise (Failure("Cannot understand command line argument " ^ (Sys.argv.(!j))))
      done;
      includingsigfile := false;
      let checkfile () =
	let c = open_in (Sys.argv.(i-1)) in
	mgcheck c;
	close_in c;
        if !pfgout then
          begin
            begin
              match !pfgtheory with
              | HF -> Printf.printf "Document 6ffc9680fbe00a58d70cdeb319f11205ed998131ce51bb96f16c7904faf74a3d\nBase set\n"
              | Mizar -> Printf.printf "Document 5ab3df7b0b4ef20889de0517a318df8746940971ad9b2021e54c820eb9e74dce\nBase set\n"
              | HOAS -> Printf.printf "Document 513140056e2032628f48d11e221efe29892e9a03a661d3b691793524a5176ede\nBase syn\n"
            end;
            List.iter
              (fun i ->
                match i with
                | PfgParam(xhv,x,agtp) ->
                   if not (pfg_prim_id_p xhv) then
                     begin
                       try
                         let pfghv = Hashtbl.find pfgtmhh xhv in
                         Printf.printf "Param %s %s : %s\n" (Hash.hashval_hexstring pfghv) x (tp_pfg_str agtp);
                         Hashtbl.add pfgtmh xhv x
                       with Not_found ->
                         Printf.printf "%% ERROR: No pfg id for %s [%s] obj\n" x xhv
                     end
                | PfgDef(xhv,x,a,m) ->
                   begin
                     tm_pfg_decl pfgdelta pfgtmph m;
                     Printf.printf "Def %s : %s\n := %s\n" x (tp_pfg_str a) (tm_pfg_str m);
                     Hashtbl.add pfgtmh xhv x
                   end
                | PfgKnown(xhv,x,p) ->
                   begin
                     tm_pfg_decl pfgdelta pfgtmph p;
                     Printf.printf "Known %s : %s\n" x (tm_pfg_str p);
	             Hashtbl.add pfgknh xhv x;
                     Hashtbl.add pfgknph xhv p;
                   end
                | PfgConj(xhv,x,p) ->
                   begin
                     tm_pfg_decl pfgdelta pfgtmph p;
                     Printf.printf "Conj %s : %s\n" x (tm_pfg_str p);
	             Hashtbl.add pfgknh xhv x;
                   end
                | PfgThm(xhv,x,p,d) ->
                   begin
                     tm_pfg_decl pfgdelta pfgtmph p;
                     pf_pfg_decl pfgdelta pfgtmph pfgknph d;
                     Printf.printf "Thm %s : %s\n := %s\n" x (tm_pfg_str p) (pf_pfg_str d);
	             Hashtbl.add pfgknh xhv x;
                     Hashtbl.add pfgknph xhv p;
                   end)
              (List.rev !pfgmain)
          end;
	begin
	  match !sigoutfile with
	  | Some(soc) -> close_out soc
	  | None -> ()
	end;
	begin
	  match !inchan with
	  | Some(c) -> close_in c
	  | None -> ()
	end;
      in
      begin
	match !solvesproblemfile with
	| None -> checkfile ()
	| Some probf ->
	    let p = open_in probf in
	    let c = open_in (Sys.argv.(i-1)) in
	    mgchecksolves p c;
	    close_in c;
	    close_in p;
	    if !verbosity > 1 then (Printf.printf "%s completely solves %s\n" Sys.argv.(i-1) probf; flush stdout);
      end
    end;
  begin
    match !latex with
    | Some hc ->
	close_out hc
    | None -> ()
  end;
  if !webout then
    begin
      Printf.printf "<div class='documentcorrect'>The document in its current form is correct.</div>\n";
    end;
  begin
    match !ownedoutfile with
    | Some(f) ->
       let c = open_out f in
       Hashtbl.iter (fun h _ -> Printf.fprintf c "Obj fake fake %s\n" (Hash.hashval_hexstring h)) ownedobj;
       Hashtbl.iter (fun h _ -> Printf.fprintf c "Prop fake fake %s\n" (Hash.hashval_hexstring h)) ownedprop;
       close_out c
    | None ->
       ()
  end;
  if (!reportids) then
    begin
      match !indoutfile with
      | Some(f) ->
	  let c = open_out f in
	  Hashtbl.iter (fun h a -> Printf.fprintf c "\"%s\" %s.\n" h (tp_to_str a)) indextms;
	  Hashtbl.iter (fun x _ -> Printf.fprintf c "Known \"%s\".\n" x) indexknowns;
	  close_out c
      | None ->
	  ()
    end
;;
